
# Copyright 2008-2020 Jaap Karssenberg <jaap.karssenberg@gmail.com>




import os
import re
import weakref
import logging
import threading

logger = logging.getLogger('zim.notebook')

from functools import partial

import zim.templates
import zim.formats

from zim.fs import adapt_from_oldfs
from zim.newfs import SEP, Folder, LocalFile, LocalFolder
from zim.config import INIConfigFile, String, ConfigDefinitionByClass, Boolean, Choice
from zim.errors import Error
from zim.base.naturalsort import natural_sort_key
from zim.newfs.helpers import TrashNotSupportedError
from zim.config import HierarchicDict
from zim.parse.links import link_type, is_win32_path_re
from zim.signals import ConnectorMixin, SignalEmitter, SIGNAL_NORMAL

from .info import create_valid_interwiki_key
from .operations import notebook_state, NOOP, SimpleAsyncOperation, ongoing_operation
from .page import Path, Page, PageError, HRef, HREF_REL_ABSOLUTE, HREF_REL_FLOATING, HREF_REL_RELATIVE
from .index import IndexNotFoundError, LINK_DIR_BACKWARD, ROOT_PATH

DATA_FORMAT_VERSION = (0, 4)


class NotebookConfig(INIConfigFile):
	'''Wrapper for the X{notebook.zim} file'''

	# TODO - unify this call with NotebookInfo ?

	def __init__(self, file):
		file = adapt_from_oldfs(file)
		INIConfigFile.__init__(self, file)
		if os.name == 'nt':
			endofline = 'dos'
		else:
			endofline = 'unix'

		self['Notebook'].define((
			('version', String('.'.join(map(str, DATA_FORMAT_VERSION)))),
			('name', String(file.parent().basename)),
			('interwiki', String(None)),
			('home', ConfigDefinitionByClass(Path('Home'))),
			('icon', String(None)), # XXX should be file, but resolves relative
			('document_root', String(None)), # XXX should be dir, but resolves relative
			('short_links', Boolean(False)),
			('shared', Boolean(True)),
			('paste_image_template', String('pasted_image_%y%m%d')),
			('endofline', Choice(endofline, {'dos', 'unix'})),
			('disable_trash', Boolean(False)),
			('default_file_format', String('zim-wiki')),
			('default_file_extension', String('.txt')),
			('notebook_layout', String('files')),
		))


def _resolve_relative_config(dir, config):
	# Some code shared between Notebook and NotebookInfo
	dir = adapt_from_oldfs(dir)

	# Resolve icon, can be relative
	icon = config.get('icon')
	if icon:
		icon = LocalFile(dir.get_abspath(icon))

	# Resolve document_root, can also be relative
	document_root = config.get('document_root')
	if document_root:
		document_root = LocalFolder(dir.get_abspath(document_root))

	return icon, document_root


def _iswritable(dir):
	if os.name == 'nt':
		# Test access - (iswritable turns out to be unreliable for folders on windows..)
		f = dir.file('.zim.tmp')
		try:
			f.write('Test')
			f.remove(cleanup=False)
		except:
			return False
		else:
			return True
	else:
		return dir.iswritable()


def _cache_dir_for_dir(dir):
	# Consider using md5 for path name here, like thumbnail spec
	from zim.config import XDG_CACHE_HOME

	if os.name == 'nt':
		path = 'notebook-' + dir.path.replace('\\', '_').replace(':', '').strip('_')
	else:
		path = 'notebook-' + dir.path.replace('/', '_').strip('_')

	return XDG_CACHE_HOME.folder(('zim', path))


class PageNotFoundError(PageError):
	_msg = _('No such page: %s') # T: message for PageNotFoundError


class PageNotAllowedError(PageNotFoundError):
	_msg = _('Page not allowed: %s') # T: message for PageNotAllowedError
	description = _('This page name cannot be used due to technical limitations of the storage')
			# T: description for PageNotAllowedError


class PageNotAvailableError(PageNotFoundError):
	_msg = _('Page not available: %s') # T: message for PageNotAvailableError
	description = _('This page name cannot be used due to a conflicting file in the storage')
			# T: description for PageNotAvailableError

	def __init__(self, path, file):
		PageError.__init__(self, path)
		self.file = file


class PageExistsError(PageError):
	_msg = _('Page already exists: %s') # T: message for PageExistsError


class IndexNotUptodateError(Error):
	pass # TODO description here?


def assert_index_uptodate(method):
	def wrapper(notebook, *arg, **kwarg):
		if not notebook.index.is_uptodate:
			raise IndexNotUptodateError('Index not up to date')
		return method(notebook, *arg, **kwarg)

	return wrapper


_NOTEBOOK_CACHE = weakref.WeakValueDictionary()


from zim.plugins import ExtensionBase, extendable

class NotebookExtension(ExtensionBase):
	'''Base class for extending the notebook

	@ivar notebook: the L{Notebook} object
	'''

	def __init__(self, plugin, notebook):
		ExtensionBase.__init__(self, plugin, notebook)
		self.notebook = notebook


@extendable(NotebookExtension)
class Notebook(ConnectorMixin, SignalEmitter):
	'''Main class to access a notebook

	This class defines an API that proxies between backend L{zim.stores}
	and L{Index} objects on the one hand and the user interface on the
	other hand. (See L{module docs<zim.notebook>} for more explanation.)

	@signal: C{store-page (page)}: emitted before actually storing the page
	@signal: C{stored-page (page)}: emitted after storing the page
	@signal: C{move-page (oldpath, newpath)}: emitted before
	actually moving a page
	@signal: C{moved-page (oldpath, newpath)}: emitted after
	moving the page
	@signal: C{delete-page (path)}: emitted before deleting a page
	@signal: C{deleted-page (path)}: emitted after deleting a page
	means that the preferences need to be loaded again as well
	@signal: C{suggest-link (path, text)}: hook that is called when trying
	to resolve links
	@signal: C{get-page-template (path)}: emitted before
	when a template for a new page is requested, intended for plugins that
	want to customize a namespace
	@signal: C{init-page-template (path, template)}: emitted before
	evaluating a template for a new page, intended for plugins that want
	to extend page templates

	@ivar name: The name of the notebook (string)
	@ivar icon: The path for the notebook icon (if any)
	# FIXME should be L{File} object
	@ivar document_root: The L{Folder} object for the X{document root} (if any)
	@ivar dir: Optional L{Folder} object for the X{notebook folder}
	@ivar file: Optional L{File} object for the X{notebook file}
	@ivar cache_dir: A L{Folder} object for the folder used to cache notebook state
	@ivar config: A L{SectionedConfigDict} for the notebook config
	(the C{X{notebook.zim}} config file in the notebook folder)
	@ivar index: The L{Index} object used by the notebook
	'''

	# define signals we want to use - (closure type, return type and arg types)
	__signals__ = {
		'store-page': (SIGNAL_NORMAL, None, (object,)),
		'stored-page': (SIGNAL_NORMAL, None, (object,)),
		'move-page': (SIGNAL_NORMAL, None, (object, object)),
		'moved-page': (SIGNAL_NORMAL, None, (object, object)),
		'delete-page': (SIGNAL_NORMAL, None, (object,)),
		'deleted-page': (SIGNAL_NORMAL, None, (object,)),
		'page-info-changed': (SIGNAL_NORMAL, None, (object,)),
		'get-page-template': (SIGNAL_NORMAL, str, (object,)),
		'init-page-template': (SIGNAL_NORMAL, None, (object, object)),

		# Hooks
		'suggest-link': (SIGNAL_NORMAL, object, (object, object)),
	}

	@classmethod
	def new_from_dir(klass, dir):
		'''Constructor to create a notebook based on a specific
		file system location.
		Since the file system is an external resource, this method
		will return unique objects per location and keep (weak)
		references for re-use.

		@param dir: a L{Folder} object
		@returns: a L{Notebook} object
		'''
		dir = adapt_from_oldfs(dir)
		assert isinstance(dir, LocalFolder)

		nb = _NOTEBOOK_CACHE.get(dir.uri)
		if nb:
			return nb

		from .index import Index
		from .layout import FilesLayout

		config = NotebookConfig(dir.file('notebook.zim'))

		if config['Notebook']['shared']:
			cache_dir = _cache_dir_for_dir(dir)
		else:
			cache_dir = dir.folder('.zim')
			cache_dir.touch()
			if not (cache_dir.exists() and _iswritable(cache_dir)):
				cache_dir = _cache_dir_for_dir(dir)

		folder = LocalFolder(dir.path)
		if config['Notebook']['notebook_layout'] == 'files':
			layout = FilesLayout(
				folder,
				config['Notebook']['endofline'],
				config['Notebook']['default_file_format'],
				config['Notebook']['default_file_extension']
			)
		else:
			raise ValueError('Unkonwn notebook layout: %s' % config['Notebook']['notebook_layout'])

		cache_dir.touch() # must exist for index to work
		index = Index(cache_dir.file('index.db').path, layout)

		nb = klass(cache_dir, config, folder, layout, index)
		_NOTEBOOK_CACHE[dir.uri] = nb
		return nb

	def __init__(self, cache_dir, config, folder, layout, index):
		'''Constructor
		@param cache_dir: a L{Folder} object used for caching the notebook state
		@param config: a L{NotebookConfig} object
		@param folder: a L{Folder} object for the notebook location
		@param layout: a L{NotebookLayout} object
		@param index: an L{Index} object
		'''
		self.folder = folder
		self.cache_dir = cache_dir
		self.state = INIConfigFile(cache_dir.file('state.conf'))
		self.config = config
		self.properties = config['Notebook']
		self.layout = layout
		self.index = index
		self._operation_check = NOOP

		self.readonly = not _iswritable(folder)

		if self.readonly:
			logger.info('Notebook read-only: %s', folder.path)

		self._page_cache = weakref.WeakValueDictionary()

		self.name = None
		self.icon = None
		self.document_root = None
		self.interwiki = None

		if folder.watcher is None:
			from zim.newfs.helpers import FileTreeWatcher
			folder.watcher = FileTreeWatcher()

		from .index import PagesView, LinksView, TagsView
		self.pages = PagesView.new_from_index(self.index)
		self.links = LinksView.new_from_index(self.index)
		self.tags = TagsView.new_from_index(self.index)

		def on_page_row_changed(o, row, oldrow):
			if row['name'] in self._page_cache:
				self._page_cache[row['name']].haschildren = row['n_children'] > 0
				self.emit('page-info-changed', self._page_cache[row['name']])

		def on_page_row_deleted(o, row):
			if row['name'] in self._page_cache:
				self._page_cache[row['name']].haschildren = False
				self.emit('page-info-changed', self._page_cache[row['name']])

		self.index.update_iter.pages.connect('page-row-changed', on_page_row_changed)
		self.index.update_iter.pages.connect('page-row-deleted', on_page_row_deleted)

		self.connectto(self.properties, 'changed', self.on_properties_changed)
		self.on_properties_changed(self.properties)

	def __repr__(self):
		return '<%s: %s>' % (self.__class__.__name__, self.name)

	def _reload_pages_in_cache(self, path):
		p = path.name
		ns = path.name + ':'
		for name, page in self._page_cache.items():
			if name == p or name.startswith(ns):
				if page.modified:
					logger.error('Page with unsaved changes in cache while modifying notebook')
				else:
					page.reload_textbuffer()
					# "page.haschildren" may also have changed, will be updated
					# by signal handlers for index

	@property
	def uri(self):
		'''Returns a file:// uri for this notebook that can be opened by zim'''
		return self.layout.root.uri

	@property
	def info(self):
		'''The L{NotebookInfo} object for this notebook'''
		try:
			uri = self.uri
		except AssertionError:
			uri = None
		from . import NotebookInfo
		return NotebookInfo(uri, **self.config['Notebook'])

	def on_properties_changed(self, properties):
		dir = self.layout.root

		self.name = properties['name'] or self.folder.basename
		icon, document_root = _resolve_relative_config(dir, properties)
		if icon:
			self.icon = icon.path # FIXME rewrite to use File object
		else:
			self.icon = None
		self.document_root = document_root

		self.interwiki = create_valid_interwiki_key(properties['interwiki'] or self.name)

	def suggest_link(self, source, word):
		'''Suggest a link Path for 'word' or return None if no suggestion is
		found. By default we do not do any suggestion but plugins can
		register handlers to add suggestions using the 'C{suggest-link}'
		signal.
		'''
		return self.emit_return_first('suggest-link', source, word)

	def get_page(self, path):
		'''Get a L{Page} object for a given path

		Typically a Page object will be returned even when the page
		does not exist. In this case the C{hascontent} attribute of
		the Page will be C{False} and C{get_parsetree()} will return
		C{None}. This means that you do not have to create a page
		explicitly, just get the Page object and store it with new
		content (if it is not read-only of course).

		However in some cases this method will return C{None}. This
		means that not only does the page not exist, but also that it
		can not be created. This should only occur for certain special
		pages and depends on the store implementation.

		@param path: a L{Path} object
		@returns: a L{Page} object or C{None}
		'''
		# As a special case, using an invalid page as the argument should
		# return a valid page object.
		assert isinstance(path, Path)
		if path.name in self._page_cache:
			page = self._page_cache[path.name]
			assert isinstance(page, Page)
			page.check_source_changed()
			return page
		else:
			file, folder = self.layout.map_page(path)
			if file.exists() and not self.layout.is_source_file(file):
				raise PageNotAvailableError(path, file)

			folder = self.layout.get_attachments_folder(path)
			format = self.layout.get_format(file)
			page = Page(path, False, file, folder, format)
			if self.readonly:
				page._readonly = True # XXX
			try:
				indexpath = self.pages.lookup_by_pagename(path)
			except IndexNotFoundError:
				pass
				# TODO trigger indexer here if page exists !
			else:
				if indexpath and indexpath.haschildren:
					page.haschildren = True
				# page might be the parent of a placeholder, in that case
				# the index knows it has children, but the store does not

			# TODO - set haschildren if page maps to a store namespace
			self._page_cache[path.name] = page
			return page

	def get_new_page(self, path):
		'''Like get_page() but guarantees the page does not yet exist
		by adding a number to the name to make it unique.

		This method is intended for cases where e.g. a automatic script
		wants to store a new page without user interaction. Conflicts
		are resolved automatically by appending a number to the name
		if the page already exists. Be aware that the resulting Page
		object may not match the given Path object because of this.

		@param path: a L{Path} object
		@returns: a L{Page} object
		'''
		i = 0
		base = path.name
		while True:
			try:
				page = self.get_page(path)
			except PageNotAvailableError:
				pass
			else:
				if not (page.hascontent or page.haschildren):
					return page
			finally:
				i += 1
				path = Path(base + ' %i' % i)

	def get_home_page(self):
		'''Returns a L{Page} object for the home page'''
		return self.get_page(self.config['Notebook']['home'])

	@notebook_state
	def store_page(self, page):
		'''Save the data from the page in the storage backend

		@param page: a L{Page} object
		@emits: store-page before storing the page
		@emits: stored-page on success
		'''
		logger.debug('Store page: %s', page)
		self.emit('store-page', page)
		page._store()
		file, folder = self.layout.map_page(page)
		self.index.update_file(file)
		page.set_modified(False)
		self.emit('stored-page', page)

	@notebook_state
	def store_page_async(self, page, parsetree):
		logger.debug('Store page in background: %s', page)
		self.emit('store-page', page)
		error = threading.Event()
		thread = threading.Thread(
			target=partial(self._store_page_async_thread_main, page, parsetree, error)
		)
		thread.start()
		pre_modified = page.modified
		op = SimpleAsyncOperation(
			notebook=self,
			message='Store page in progress',
			thread=thread,
			post_handler=partial(self._store_page_async_finished, page, error, pre_modified)
		)
		op.error_event = error
		op.run_on_idle()
		return op

	def _store_page_async_thread_main(self, page, parsetree, error):
		try:
			page._store_tree(parsetree)
		except:
			error.set()
			logger.exception('Error in background save')

	def _store_page_async_finished(self, page, error, pre_modified):
		if not error.is_set():
			file, folder = self.layout.map_page(page)
			self.index.update_file(file)
			if page.modified == pre_modified:
				# HACK: Checking modified state protects against race condition
				# in async store. Works because pageview sets "page.modified"
				# to a counter rather than a boolean
				page.set_modified(False)
				self.emit('stored-page', page)

	def wait_for_store_page_async(self):
		op = ongoing_operation(self)
		if isinstance(op, SimpleAsyncOperation):
			op()

	def move_page(self, path, newpath, update_links=True, update_heading=False):
		'''Move and/or rename a page in the notebook

		@param path: a L{Path} object for the old/current page name
		@param newpath: a L{Path} object for the new page name
		@param update_links: if C{True} all links B{from} and B{to} this
		page and any of it's children will be updated to reflect the
		new page name
		@param update_heading: if C{True} the heading of the page will be
		changed to the basename of the new path

		The original page C{path} does not have to exist, in this case
		only the link update will done. This is useful to update links
		for a placeholder.

		@raises PageExistsError: if C{newpath} already exists

		@emits: move-page before the move
		@emits: moved-page after successfull move
		'''
		for p in self.move_page_iter(path, newpath, update_links, update_heading):
			pass

	@assert_index_uptodate
	@notebook_state
	def move_page_iter(self, path, newpath, update_links=True, update_heading=False):
		'''Like L{move_page()} but yields pages that are being updated
		if C{update_links} is C{True}
		'''
		logger.debug('Move page %s to %s', path, newpath)

		self.emit('move-page', path, newpath)
		try:
			n_links = self.links.n_list_links_section(path, LINK_DIR_BACKWARD)
		except IndexNotFoundError:
			raise PageNotFoundError(path)

		file, folder = self.layout.map_page(path)
		if (file.exists() or folder.exists()):
			self._move_file_and_folder(path, newpath)
			self._reload_pages_in_cache(path)
			self._reload_pages_in_cache(newpath)
			self.emit('moved-page', path, newpath)

			if update_links:
				for p in self._update_links_in_moved_page(path, newpath):
					yield p

		if update_links:
			for p in self._update_links_to_moved_page(path, newpath):
				yield p

			new_n_links = self.links.n_list_links_section(newpath, LINK_DIR_BACKWARD)
			if new_n_links != n_links:
				logger.warning('Number of links after move (%i) does not match number before move (%i)', new_n_links, n_links)
			else:
				logger.debug('Number of links after move does match number before move (%i)', new_n_links)

		if update_heading:
			page = self.get_page(newpath)
			tree = page.get_parsetree()
			if not tree is None:
				tree.set_heading_text(newpath.basename)
				page.set_parsetree(tree)
				self.store_page(page)

	def _move_file_and_folder(self, path, newpath):
		file, folder = self.layout.map_page(path)
		if not (file.exists() or folder.exists()):
			raise PageNotFoundError(path)

		newfile, newfolder = self.layout.map_page(newpath)
		if file.path.lower() == newfile.path.lower():
			if newfile.isequal(file) or newfolder.isequal(folder):
				pass # renaming on case-insensitive filesystem
			elif newfile.exists() or newfolder.exists():
				raise PageExistsError(newpath)
		elif newfile.exists():
			if self.layout.is_source_file(newfile):
				raise PageExistsError(newpath)
			else:
				raise PageNotAvailableError(newpath, newfile)
		elif newfolder.exists():
			raise PageExistsError(newpath)

		# First move the dir - if it fails due to some file being locked
		# the whole move is cancelled. Chance is bigger than the other
		# way around, e.g. attachment open in external program.

		changes = []

		if folder.exists():
			if newfolder.ischild(folder):
				# special case where we want to move a page down
				# into it's own namespace
				parent = folder.parent()
				tmp = parent.new_folder(folder.basename)
				folder.moveto(tmp)
				tmp.moveto(newfolder)
			else:
				folder.moveto(newfolder)

			changes.append((folder, newfolder))

			# check if we also moved the file inadvertently
			if file.ischild(folder):
				rel = file.relpath(folder)
				movedfile = newfolder.file(rel)
				if movedfile.exists() and movedfile.path != newfile.path:
						movedfile.moveto(newfile)
						changes.append((movedfile, newfile))
			elif file.exists():
				file.moveto(newfile)
				changes.append((file, newfile))

		elif file.exists():
			file.moveto(newfile)
			changes.append((file, newfile))

		# Process index changes after all fs changes
		# more robust if anything goes wrong in index update
		for old, new in changes:
			self.index.file_moved(old, new)


	def _update_links_in_moved_page(self, oldroot, newroot):
		# Find (floating) links that originate from the moved page
		# check if they would resolve different from the old location
		seen = set()
		for link in list(self.links.list_links_section(newroot)):
			if link.source.name not in seen:
				if link.source == newroot:
					oldpath = oldroot
				else:
					oldpath = oldroot + link.source.relname(newroot)

				yield link.source
				self._update_moved_page(link.source, oldpath, newroot, oldroot)
				seen.add(link.source.name)

	def _update_moved_page(self, path, oldpath, newroot, oldroot):
		logger.debug('Updating links in page moved from %s to %s', oldpath, path)
		page = self.get_page(path)
		tree = page.get_parsetree()
		if not tree:
			return

		def replacefunc(elt):
			text = elt.attrib['href']
			if link_type(text) != 'page':
				return elt

			href = HRef.new_from_wiki_link(text)
			if href.rel == HREF_REL_RELATIVE:
				pass
			elif href.rel == HREF_REL_ABSOLUTE:
				oldtarget = self.pages.resolve_link(page, href)
				if oldtarget == oldroot:
					return self._update_link_tag(elt, page, newroot, href)
				elif oldtarget.ischild(oldroot):
					newtarget = newroot + oldtarget.relname(oldroot)
					return self._update_link_tag(elt, page, newtarget, href)
			else:
				assert href.rel == HREF_REL_FLOATING
				newtarget = self.pages.resolve_link(page, href)
				oldtarget = self.pages.resolve_link(oldpath, href)

				if oldtarget == oldroot:
					return self._update_link_tag(elt, page, newroot, href)
				elif oldtarget.ischild(oldroot) and href.names:  # make sure href has parts
					oldanchor = self.pages.resolve_link(oldpath, HRef(HREF_REL_FLOATING, href.parts()[0]))
					if oldanchor.ischild(oldroot):
						pass # oldtarget cannot be trusted
					else:
						newtarget = newroot + oldtarget.relname(oldroot)
						return self._update_link_tag(elt, page, newtarget, href)
				elif newtarget != oldtarget:
					# Redirect back to old target
					return self._update_link_tag(elt, page, oldtarget, href)

			return elt

		newtree = tree.substitute_elements((zim.formats.LINK,), replacefunc)
		page.set_parsetree(newtree)
		self.store_page(page)

	def _update_links_to_moved_page(self, oldroot, newroot):
		# 1. Check remaining placeholders, update pages causing them
		seen = set()
		try:
			oldroot = self.pages.lookup_by_pagename(oldroot)
		except IndexNotFoundError:
			pass
		else:
			for link in list(self.links.list_links_section(oldroot, LINK_DIR_BACKWARD)):
				if link.source.name not in seen:
					yield link.source
					self._move_links_in_page(link.source, oldroot, newroot)
					seen.add(link.source.name)

		# 2. Check for links that have anchor of same name as the moved page
		# and originate from a (grand)child of the parent of the moved page
		# and no longer resolve to the moved page
		parent = oldroot.parent
		for link in list(self.links.list_floating_links(oldroot.basename)):
			if link.source.name not in seen \
			and link.source.ischild(parent) \
			and not (
				link.target == newroot
				or link.target.ischild(newroot)
			):
				yield link.source
				self._move_links_in_page(link.source, oldroot, newroot)
				seen.add(link.source.name)

	def _move_links_in_page(self, path, oldroot, newroot):
		logger.debug('Updating page %s to move link from %s to %s', path, oldroot, newroot)
		page = self.get_page(path)
		tree = page.get_parsetree()
		if not tree:
			return

		def replacefunc(elt):
			text = elt.attrib['href']
			if link_type(text) != 'page':
				return elt

			href = HRef.new_from_wiki_link(text)
			target = self.pages.resolve_link(page, href)

			if target == oldroot:
				return self._update_link_tag(elt, page, newroot, href)
			elif target.ischild(oldroot):
				newtarget = newroot.child(target.relname(oldroot))
				return self._update_link_tag(elt, page, newtarget, href)

			elif href.rel == HREF_REL_FLOATING and href.names \
			and natural_sort_key(href.parts()[0]) == natural_sort_key(oldroot.basename) \
			and page.ischild(oldroot.parent):
				try:
					targetrecord = self.pages.lookup_by_pagename(target)
				except IndexNotFoundError:
					targetrecord = None # technically this is a bug, but let's be robust

				if not target.ischild(oldroot.parent) \
				or targetrecord is None or not targetrecord.exists():
					# An link that was anchored to the moved page,
					# but now resolves somewhere higher in the tree
					# Or a link that no longer resolves
					if len(href.parts()) == 1:
						return self._update_link_tag(elt, page, newroot, href)
					else:
						mynewroot = newroot.child(':'.join(href.parts()[1:]))
						return self._update_link_tag(elt, page, mynewroot, href)

			return elt

		newtree = tree.substitute_elements((zim.formats.LINK,), replacefunc)
		page.set_parsetree(newtree)
		self.store_page(page)

	def _update_link_tag(self, elt, source, target, oldhref):
		if oldhref.rel == HREF_REL_ABSOLUTE: # prefer to keep absolute links
			newhref = HRef(HREF_REL_ABSOLUTE, target.name)
		elif source == target and oldhref.anchor:
			newhref = HRef(HREF_REL_FLOATING, '', oldhref.anchor)
		else:
			newhref = self.pages.create_link(source, target)

		newhref.anchor = oldhref.anchor

		link = newhref.to_wiki_link()

		from zim.formats import TEXT
		if elt.content == [(TEXT, elt.attrib['href'])]:
			elt.content[:] = [(TEXT, link)]
		elif elt.content == [(TEXT, oldhref.short_name())]:
			# Related to 'short_links' but not checking the property here.
			elt.content[:] = [(TEXT, newhref.short_name())]  # 'Journal:2020:01:20' -> '20'

		elt.attrib['href'] = link

		return elt

	@assert_index_uptodate
	@notebook_state
	def delete_page(self, path, update_links=True):
		'''Delete a page from the notebook

		@param path: a L{Path} object
		@param update_links: if C{True} pages linking to the
		deleted page will be updated and the link are removed.

		@returns: C{True} when the page existed and was deleted,
		C{False} when the page did not exist in the first place.

		Raises an error when delete failed.

		@emits: delete-page before the actual delete
		@emits: deleted-page after successfull deletion
		'''
		existed = self._delete_page(path)

		for p in self._deleted_page(path, update_links):
			pass

		return existed

	@assert_index_uptodate
	@notebook_state
	def delete_page_iter(self, path, update_links=True):
		'''Like L{delete_page()}'''
		self._delete_page(path)

		for p in self._deleted_page(path, update_links):
			yield p

	def _delete_page(self, path):
		logger.debug('Delete page: %s', path)
		self.emit('delete-page', path)

		file, folder = self.layout.map_page(path)
		assert file.path.startswith(self.folder.path)
		assert folder.path.startswith(self.folder.path)

		if not (file.exists() or folder.exists()):
			return False
		else:
			if folder.exists():
				folder.remove_children()
				folder.remove()
			if file.exists():
				file.remove()

			self.index.update_file(file)
			self.index.update_file(folder)

			return True

	@assert_index_uptodate
	@notebook_state
	def trash_page(self, path, update_links=True):
		'''Move a page to Trash

		Like L{delete_page()} but will use the system Trash (which may
		depend on the OS we are running on). This is used in the
		interface as a more user friendly version of delete as it is
		undoable.

		@param path: a L{Path} object
		@param update_links: if C{True} pages linking to the
		deleted page will be updated and the link are removed.

		@returns: C{True} when the page existed and was deleted,
		C{False} when the page did not exist in the first place.

		Raises an error when trashing failed.

		@raises TrashNotSupportedError: if trashing is not supported by
		the storage backend or when trashing is explicitly disabled
		for this notebook.

		@emits: delete-page before the actual delete
		@emits: deleted-page after successfull deletion
		'''
		existed = self._trash_page(path)

		for p in self._deleted_page(path, update_links):
			pass

		return existed

	@assert_index_uptodate
	@notebook_state
	def trash_page_iter(self, path, update_links=True):
		'''Like L{trash_page()}'''
		self._trash_page(path)

		for p in self._deleted_page(path, update_links):
			yield p

	def _trash_page(self, path):
		from zim.newfs.helpers import TrashHelper

		logger.debug('Trash page: %s', path)

		if self.config['Notebook']['disable_trash']:
			raise TrashNotSupportedError('disable_trash is set')

		self.emit('delete-page', path)

		file, folder = self.layout.map_page(path)
		helper = TrashHelper()

		re = False
		if folder.exists():
			re = helper.trash(folder)
			if isinstance(path, Page):
				path.haschildren = False

		if file.exists():
			re = helper.trash(file) or re

		self.index.update_file(file)
		self.index.update_file(folder)

		return re

	def _deleted_page(self, path, update_links):
		self._reload_pages_in_cache(path)
		path = Path(path.name)

		if update_links:
			# remove persisting links
			try:
				indexpath = self.pages.lookup_by_pagename(path)
			except IndexNotFoundError:
				pass
			else:
				pages = set(
					l.source for l in self.links.list_links_section(path, LINK_DIR_BACKWARD))

				for p in pages:
					yield p
					page = self.get_page(p)
					self._remove_links_in_page(page, path)
					self.store_page(page)

		# let everybody know what happened
		self.emit('deleted-page', path)

	def _remove_links_in_page(self, page, path):
		logger.debug('Removing links in %s to %s', page, path)
		tree = page.get_parsetree()
		if not tree:
			return

		def replacefunc(elt):
			href = elt.attrib['href']
			type = link_type(href)
			if type != 'page':
				return elt

			hrefpath = self.pages.lookup_from_user_input(href, page)
			#~ print('LINK', hrefpath)
			if hrefpath == path \
			or hrefpath.ischild(path):
				# Replace the link by it's text
				return elt.content
			else:
				return elt

		newtree = tree.substitute_elements((zim.formats.LINK,), replacefunc)
		page.set_parsetree(newtree)

	def resolve_file(self, filename, path=None):
		'''Resolve a file or directory path relative to a page or
		Notebook

		This method is intended to lookup file links found in pages and
		turn resolve the absolute path of those files.

		File URIs and paths that start with '~/' or '~user/' are
		considered absolute paths. Also windows path names like
		'C:\\user' are recognized as absolute paths.

		Paths that starts with a '/' are taken relative to the
		to the I{document root} - this can e.g. be a parent directory
		of the notebook. Defaults to the filesystem root when no document
		root is set. (So can be relative or absolute depending on the
		notebook settings.)

		Paths starting with any other character are considered
		attachments. If C{path} is given they are resolved relative to
		the I{attachment folder} of that page, otherwise they are
		resolved relative to the I{notebook folder} - if any.

		Paths ending with a "/" or "\" are considered folders.

		The file is resolved purely based on the path, it does not have
		to exist at all. However if a folder of the name exists a L{Folder}
		object is returned instead of a file.

		@param filename: the (relative) file path or uri as string
		@param path: a L{Path} object for the page
		@returns: a L{File} or L{Folder} object.
		'''
		assert isinstance(filename, str) and filename
		try:
			file = self._resolve_abs_file(filename)
		except:
			logger.exception('Could not resolve filename as absolute path: %s', filename)
			file = None

		if file is None:
			if path:
				folder = self.get_attachments_dir(path)
			else:
				folder = self.layout.root

			file = LocalFile(folder.get_abspath(filename))

		myfolder = LocalFolder(file)
		if filename[-1] in ('/', '\\') or myfolder.exists():
			return myfolder
		else:
			return file

	def _resolve_abs_file(self, filename):
		# Code shared between notebook & export linker
		filename = filename.replace('\\', '/')
		if filename.startswith('~') or filename.startswith('file:/'):
			file = LocalFile(filename) # Note: can raise for non-local file URI
		elif filename.startswith('/'):
			if self.document_root:
				file = self.document_root.file(filename)
			else:
				file = LocalFile(filename)
		elif is_win32_path_re.match(filename):
			if not filename.startswith('/'):
				filename = '/' + filename # make absolute on Unix
			file = LocalFile(filename)
		else:
			file = None

		return file

	def relative_filepath(self, file, path=None):
		'''Get a file path relative to the notebook or page

		Intended as the counter part of L{resolve_file()}. Typically
		this function is used to present the user with readable paths or to
		shorten the paths inserted in the wiki code. It is advised to
		use file URIs for links that can not be made relative with
		this method.

		The link can be relative:
		  - to the I{document root} (link will start with "/")
		  - the attachments dir (if a C{path} is given) or the notebook
		    (links starting with "./" or "../")
		  - or the users home dir (link like "~/user/")

		Relative file paths are always given with Unix path semantics
		(so "/" even on windows). But a leading "/" does not mean the
		path is absolute, but rather that it is relative to the
		X{document root}.

		@param file: L{File} object we want to link
		@keyword path: L{Path} object for the page where we want to
		link this file

		@returns: relative file path as string, or C{None} when no
		relative path was found
		'''
		file = adapt_from_oldfs(file)
		if not file.islocal:
			return None

		notebook_root = self.layout.root
		document_root = LocalFolder(self.document_root.path) if self.document_root else None# XXX

		rootdir = '/'
		mydir = '.' + SEP
		updir = '..' + SEP
		postfix = SEP if isinstance(file, Folder) else ''

		# Look within the notebook
		if path:
			attachments_dir = self.get_attachments_dir(path)

			if file.ischild(attachments_dir):
				return mydir + file.relpath(attachments_dir) + postfix
			elif document_root and notebook_root \
			and document_root.ischild(notebook_root) \
			and file.ischild(document_root) \
			and not attachments_dir.ischild(document_root):
				# special case when document root is below notebook root
				# the case where document_root == attachment_folder is
				# already caught by above if clause
				return rootdir + file.relpath(document_root) + postfix
			elif notebook_root \
			and file.ischild(notebook_root) \
			and attachments_dir.ischild(notebook_root):
				parent = file.commonparent(attachments_dir)
				uppath = attachments_dir.relpath(parent)
				downpath = file.relpath(parent)
				up = 1 + uppath.replace('\\', '/').count('/')
				return updir * up + downpath + postfix
		else:
			if document_root and notebook_root \
			and document_root.ischild(notebook_root) \
			and file.ischild(document_root):
				# special case when document root is below notebook root
				return rootdir + file.relpath(document_root) + postfix
			elif notebook_root and file.ischild(notebook_root):
				return mydir + file.relpath(notebook_root) + postfix

		# If that fails look for global folders
		if document_root and file.ischild(document_root):
			return rootdir + file.relpath(document_root) + postfix

		# Finally check HOME or give up
		path = file.userpath + postfix
		return path if path.startswith('~') else None

	def get_attachments_dir(self, path):
		'''Get the X{attachment folder} for a specific page

		@param path: a L{Path} object
		@returns: a L{Folder} object or C{None}

		Always returns an object when the page can have an attachment
		folder, even when the folder does not (yet) exist. However when
		C{None} is returned the store implementation does not support
		an attachments folder for this page.
		'''
		return self.layout.get_attachments_folder(path)

	def get_template(self, path, context=None):
		'''Get a template for the intial text on new pages
		@param path: a L{Path} object
		@param context: optional dict with additional context parameters
		@returns: a L{ParseTree} object
		'''
		# FIXME hardcoded that template must be wiki format

		template = self.get_page_template_name(path)
		logger.debug('Got page template \'%s\' for %s', template, path)
		template = zim.templates.get_template('wiki', template)
		return self.eval_new_page_template(path, template, context)

	def get_page_template_name(self, path=None):
		'''Returns the name of the template to use for a new page.
		(To get the contents of the template directly, see L{get_template()})
		'''
		return self.emit_return_first('get-page-template', path or Path(':')) or 'Default'

	def eval_new_page_template(self, path, template, context=None):
		lines = []
		mycontext = {
			'page': {
				'name': path.name,
				'basename': path.basename,
				'section': path.namespace,
				'namespace': path.namespace, # backward compat
			}
		}
		if context:
			mycontext.update(context)
		self.emit('init-page-template', path, template) # plugin hook
		template.process(lines, mycontext)

		parser = zim.formats.get_parser('wiki')
		return parser.parse(lines)




# Copyright 2008,2015,2023-2025 Jaap Karssenberg <jaap.karssenberg@gmail.com>


import re
import locale
import logging

logger = logging.getLogger('zim.plugins.spell')


from zim.plugins import PluginClass
from zim.signals import SIGNAL_AFTER, ConnectorMixin
from zim.actions import toggle_action

from zim.gui.pageview import PageViewExtension
from zim.gui.widgets import ErrorDialog

import gi

# Try which of the bindings is available

gtkspellcheck = None
gtkspell = None
Gspell = None


try:
	gi.require_version('Gtk','3.0') # see issue #2301
	import gtkspellcheck
except:
	gtkspellcheck = None

if not gtkspellcheck:
	try:
		gi.require_version('Gspell', '1')
		from gi.repository import Gspell
		
		langs = Gspell.language_get_available()
		#for lang in langs:
		#	logger.debug('%s (%s) dict available', lang.get_name(), lang.get_code())
		if not langs:
			Gspell = None
	except:
		Gspell = None

if not Gspell:
	try:
		gi.require_version('GtkSpell', '3.0')
		from gi.repository import GtkSpell as gtkspell
	except:
		gtkspell = None



# Hotfix for robustness of loading languages in gtkspellcheck
# try to be robust for future versions breaking this or not needing it
# See https://github.com/koehlma/pygtkspellcheck/issues/22
#
# gtkspellchecker 5 has removed "pylocales", so if not present the fix is no
# longer working. Not sure if it is still needed with this release.
try:
	import pylocales
except ImportError:
	pylocales = None

if gtkspellcheck and pylocales \
and hasattr(gtkspellcheck.SpellChecker, '_LanguageList') \
and hasattr(gtkspellcheck.SpellChecker._LanguageList, 'from_broker'):
	from pylocales import code_to_name

	orig_from_broker = gtkspellcheck.SpellChecker._LanguageList.from_broker

	@classmethod
	def new_from_broker(cls, broker):
		try:
			return orig_from_broker(broker)
		except:
			lang = []
			for language in broker.list_languages():
				try:
					lang.append((language, code_to_name(language)))
				except:
					logger.exception('While loading language for: %s', language)

			return cls(sorted(lang, key=lambda language: language[1]))

	gtkspellcheck.SpellChecker._LanguageList.from_broker = new_from_broker
#####



# Silence gtkspellcheck logging, it is very verbose
mylogger = logging.getLogger('gtkspellcheck')
mylogger.setLevel(logging.INFO)
#####


def choose_adapter_cls():
	if gtkspellcheck:
		version = tuple(
			map(int, re.findall(r'\d+', gtkspellcheck.__version__))
		)
		if version >= (4, 0, 3):
			return GtkspellcheckAdapter
		else:
			logger.warning(
				'Using gtkspellcheck %s. Versions before 4.0.3 might cause memory leak.',
				gtkspellcheck.__version__
			)
			return OldGtkspellcheckAdapter
	elif Gspell:
		return GspellAdapter
	else:
		return GtkspellAdapter


class SpellPlugin(PluginClass):

	plugin_info = {
		'name': _('Spell Checker'), # T: plugin name
		'description': _('''\
Adds spell checking support using 
gtkspellchecker, Gspell, or gtkspell libraries.

This is a core plugin shipping with zim.
'''), # T: plugin description
		'author': 'Jaap Karssenberg',
		'help': 'Plugins:Spell Checker',
	}

	plugin_notebook_properties = (
		('language', 'string', 'Default Language', ''),
	)

	@classmethod
	def check_dependencies(klass):
		return any((gtkspellcheck, Gspell, gtkspell)), [
			('gtkspellcheck', not gtkspellcheck is None, True),
			('Gspell', not Gspell is None, True),
			('gtkspell', not gtkspell is None, True)
		]


class SpellPageViewExtension(PageViewExtension):

	def __init__(self, plugin, pageview):
		PageViewExtension.__init__(self, plugin, pageview)
		self._adapter_cls = choose_adapter_cls()
		self.checker = None

		self.uistate.setdefault('active', False)
		self.toggle_spellcheck(self.uistate['active'])

		self.properties = self.plugin.notebook_properties(self.pageview.notebook)
		self.on_properties_changed(self.properties)
		self.connectto(self.properties, 'changed', self.on_properties_changed)

		self.connectto(self.pageview, 'page-changed', order=SIGNAL_AFTER)

	def on_properties_changed(self, properties):
		if self.checker:
			self.checker.teardown()

		textview = self.pageview.textview
		lang = self.properties['language'] or locale.getdefaultlocale()[0]
		logger.debug('Spellcheck language: %s', lang)
		try:
			self.checker = self._adapter_cls(textview, lang)
			if self.uistate['active']:
				self.checker.enable()
			else:
				self.checker.disable()
		except:
			ErrorDialog(self.pageview, (
				_('Could not load spell checking'),
					# T: error message
				_('This could mean you don\'t have the proper\ndictionaries installed')
					# T: error message explanation
			)).run()

	def on_page_changed(self, pageview, page):
		# A new buffer may be initialized, but it could also be an existing buffer linked to page
		if self.checker:
			self.checker.on_buffer_changed(pageview.textview, self.uistate['active'])

	@toggle_action(_('Check _spelling'), accelerator='F7') # T: menu item
	def toggle_spellcheck(self, active):
		if self.checker:
			if active:
				self.checker.enable()
			else:
				self.checker.disable()

		self.uistate['active'] = active

	def teardown(self):
		if self.checker:
			self.checker.teardown()
		self.checker = None


class AdapterBase(ConnectorMixin, object):

	def __init__(self, textview, language):
		'''Contructor
		@param textview: a C{Gtk.TextView} to apply spellchecking to
		@param langage: a language code as string
		'''
		raise NotImplementedError

	def on_buffer_changed(self, textview, active):
		'''Callback for when the C{Gtk.TextBuffer} in the C{Gtk.TextView} was changed
		@param textview: a C{Gtk.TextView} to apply spellchecking to
		@param active: whether spellchecking is toggled on or off
		'''
		pass

	def enable(self):
		'''Toggle on the spell check behavior'''
		raise NotImplementedError

	def disable(self):
		'''Toggle off the spell check behavior'''
		raise NotImplementedError

	def teardown(self):
		'''Clean up object state before destruction'''
		raise NotImplementedError


class GtkspellcheckAdapter(AdapterBase):

	def __init__(self, textview, lang):
		self._lang = lang
		self._textview = textview

		self._clean_tag_table() # Just in case
		self._checker = gtkspellcheck.SpellChecker(self._textview, self._lang)

	def on_buffer_changed(self, textview, active):
		# Check whether buffer was initialized already by inspecting tag table
		if not self._check_tag_table():
			self._checker.buffer_initialize()

	def enable(self):
		self._checker.enable()

	def disable(self):
		self._checker.disable()

	def teardown(self):
		self._checker.disable()
		self._clean_tag_table()

	def _check_tag_table(self):
		tags = []

		def filter_spell_tags(t):
			name = t.get_property('name')
			if name and name.startswith('gtkspellchecker'):
				tags.append(t)

		table = self._textview.get_buffer().get_tag_table()
		table.foreach(filter_spell_tags)
		return tags

	def _clean_tag_table(self):
		## cleanup tag table - else next loading will fail
		table = self._textview.get_buffer().get_tag_table()
		for tag in self._check_tag_table():
			table.remove(tag)


class OldGtkspellcheckAdapter(GtkspellcheckAdapter):

	def on_buffer_changed(self, textview, active):
		# Check whether buffer was initialized already by inspecting tag table
		# wanted to use checker.buffer_initialize() here,
		# but gives issue, see https://github.com/koehlma/pygtkspellcheck/issues/24
		# So, just re-initialize
		if not self._check_tag_table():
			self._checker.disable()
			self._clean_tag_table()
			self._checker = gtkspellcheck.SpellChecker(self._textview, self._lang)
			if active:
				self._checker.enable()


class GtkspellAdapter(AdapterBase):

	def __init__(self, textview, lang):
		self._lang = lang
		self._textview = textview
		self._checker = gtkspell.Checker()
		self._checker.set_language(self._lang)

	def enable(self):
		self._checker.attach(self._textview)

	def disable(self):
		self._checker.detach()

	def teardown(self):
		self._checker.detach()


class GspellAdapter(AdapterBase):

	def __init__(self, textview, lang):
		gspell_language = Gspell.language_lookup(lang)
		self._checker = Gspell.Checker.new(gspell_language)
		self.on_buffer_changed(textview, False)
		self._gspell_view = Gspell.TextView.get_from_gtk_text_view(textview)
		self.enable()

	def on_buffer_changed(self, textview, active):
		buffer = Gspell.TextBuffer.get_from_gtk_text_buffer(textview.get_buffer())
		buffer.set_spell_checker(self._checker)

	def enable(self):
		self._gspell_view.set_inline_spell_checking(True)
		self._gspell_view.set_enable_language_menu(True)
		
	def disable(self):
		self._gspell_view.set_inline_spell_checking(False)
		self._gspell_view.set_enable_language_menu(False)

	def teardown(self):
		self.disable() # No real teardown (?)




# Copyright 2008-2023 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module contains helper functions to deal with images in the user interface'''

import logging

from gi.repository import GLib, GdkPixbuf

from zim.newfs import LocalFile

logger = logging.getLogger('zim.gui.base.images')


def supports_image_format(fmt):
	'''Check if an image format is supported by GDK.
	'''
	return fmt in (f.get_name() for f in GdkPixbuf.Pixbuf.get_formats())


def image_file_load_pixels(file: LocalFile, width_override=-1, height_override=-1, fail_silent=False) -> GdkPixbuf.Pixbuf:
	"""
	Replacement for GdkPixbuf.Pixbuf.new_from_file_at_size(file.path, w, h)
	Does it's best to rotate the image to the right orientation.
	When file does not exist or fails to load, this throws exceptions.
	"""

	if not file.exists():
		# if the file does not exist, no need to make the effort of trying to read it
		raise FileNotFoundError(file.path)

	need_switch_to_fallback = True
	try:
		pixbuf = GdkPixbuf.Pixbuf.new_from_file(file.path)
	except GLib.GError:
		if not fail_silent:
			logger.debug(f'GTK failed to read image, using Pillow fallback: {file.path}')
	else:
		need_switch_to_fallback = False

	if need_switch_to_fallback:
		try:
			# load Pillow only if necessary
			# noinspection PyUnresolvedReferences
			from PIL import Image, UnidentifiedImageError
		except ImportError:
			pass
		else:
			try:
				with Image.open(file.path) as image:
					pixbuf = _convert_pillow_image_to_pixbuf(image)
			except UnidentifiedImageError:
				if not fail_silent:
					logger.debug(f'Pillow failed to read image: {file.path}')
			else:
				need_switch_to_fallback = False

	if need_switch_to_fallback:
		error_message = f'No available fallback for load this image: {file.path}'
		if not fail_silent:
			logger.debug(error_message)
		raise TypeError(error_message)

	# Let's try to find and remember the orientation before scaling,
	# 	because we lose metadata when changing images.
	orientation = None
	mimetype = file.mimetype()
	if mimetype in {'image/jpeg', 'image/tiff'}:
		# Gtk can detect orientation in jpeg|tiff images only
		# See docs: https://docs.gtk.org/gdk-pixbuf/method.Pixbuf.get_option.html#description
		orientation = pixbuf.get_option('orientation')
	if mimetype in {'image/webp', 'image/png'}:
		# if possible, we will find orientation of the image using Pillow,
		# 	if it is not available, we will display image it as is.
		try:
			# noinspection PyUnresolvedReferences
			from PIL import Image, UnidentifiedImageError, __version__ as pillow_version_string
		except ImportError:
			pass
		else:
			pillow_version = tuple(map(int, pillow_version_string.split('.')))
			try:
				with Image.open(file.path) as image:
					if pillow_version >= (6, 0, 0):
						# https://pillow.readthedocs.io/en/stable/releasenotes/6.0.0.html#added-exif-class
						exif = image.getexif()
					else:
						# noinspection PyUnresolvedReferences,PyProtectedMember
						exif = image._getexif()  # noqa: WPS437
						if not exif:
							exif = {}  # noqa: WPS437
			except UnidentifiedImageError:
				logger.debug(f'Pillow failed to read image: {file.path}')
			else:
				orientation_tag_id = 274
				orientation = exif.get(orientation_tag_id)
	orientation = int(orientation) if orientation else 1

	w, h = pixbuf.get_width(), pixbuf.get_height()
	b_size_override = width_override > 0 or height_override > 0
	if b_size_override and (width_override <= 0 or height_override <= 0):
		if orientation in {5, 6, 7, 8}:
			w, h = h, w
			if height_override <= 0:
				height_override = int(h * width_override / w)
			else:
				width_override = int(w * height_override / h)
			width_override, height_override = height_override, width_override
		else:
			if height_override <= 0:
				height_override = int(h * width_override / w)
			else:
				width_override = int(w * height_override / h)

	if b_size_override:
		# do not use new_from_file_at_size() here due to bug in Gtk for GIF images, see issue #1563
		pixbuf = pixbuf.scale_simple(width_override, height_override, GdkPixbuf.InterpType.BILINEAR)

	pixbuf.set_option('orientation', f'{orientation}')
	#      ^^^ be sure to attach the tag for magic from stdlib in next line.
	return GdkPixbuf.Pixbuf.apply_embedded_orientation(pixbuf)


def image_file_get_dimensions(file_path):
	"""
	Replacement for GdkPixbuf.Pixbuf.get_file_info
	@return (width, height) in pixels
		or None if file does not exist or failed to load
	"""

	# Let GTK try reading the file
	_, width, height = GdkPixbuf.Pixbuf.get_file_info(file_path)
	if width > 0 and height > 0:
		return (width, height)

	# Fallback to Pillow
	try:
		from PIL import Image # load Pillow only if necessary
		with Image.open(file_path) as img_pil:
			return (img_pil.width, img_pil.height)
	except:
		raise AssertionError('Could not get size for: %s' % file_path)


def _convert_pillow_image_to_pixbuf(image) -> GdkPixbuf.Pixbuf:
	# check if there is an alpha channel
	if image.mode == 'RGB':
		has_alpha = False
	elif image.mode == 'RGBA':
		has_alpha = True
	else:
		raise ValueError('Pixel format {fmt} can not be converted to Pixbuf for image {image}'.format(
			fmt=image.mode, image=image,
		))

	# convert to GTK pixbuf
	data_gtk = GLib.Bytes.new_take(image.tobytes())

	return GdkPixbuf.Pixbuf.new_from_bytes(
		data=data_gtk,
		colorspace=GdkPixbuf.Colorspace.RGB,
		has_alpha=has_alpha,
		# GTK docs: "Currently only RGB images with 8 bits per sample are supported"
		# https://docs.gtk.org/gdk-pixbuf/ctor.Pixbuf.new_from_bytes.html#description
		bits_per_sample=8,
		width=image.width,
		height=image.height,
		rowstride=image.width * (4 if has_alpha else 3),
	)




# Copyright 2022 Jaap Karssenberg <jaap.karssenberg@gmail.com>


import logging

logger = logging.getLogger('zim.plugins.toolbar')

from gi.repository import Gtk

from zim.plugins import PluginClass
from zim.signals import SignalHandler
from zim.gui.pageview import PageViewExtension
from zim.plugins.tasklist.gui import TaskListWindowExtension
from zim.gui.widgets import TOP, POSITIONS


STYLES = (
	('ICONS', _('Icons')), # T: toolbar style
	('TEXT', _('Text')), # T: toolbar style
	('BOTH', _('Icons & Text')), # T: toolbar style
	('BOTH_HORIZ', _('Icons & Text horizontal')), # T: toolbar style
)

def _get_style(key):
	return getattr(Gtk.ToolbarStyle, key)


SIZES = (
	('SMALL_TOOLBAR', _('Small')), # T: Toolbar size small
	('LARGE_TOOLBAR', _('Large')), # T: Toolbar size large
)

def _get_size(key):
	return getattr(Gtk.IconSize, key)


class ToolBarPlugin(PluginClass):

	plugin_info = {
		'name': _('Tool Bar'), # T: plugin name
		'description': _('''\
This plugin allows to customize the toolbar in the main window.
'''), # T: plugin description
		'author': 'Jaap Karssenberg',
		'help': 'Plugins:ToolBar',
	}

	plugin_preferences = (
		# key, type, label, default
		('show_toolbar', 'bool', _('Show toolbar'), True),
			# T: option for plugin preferences
		('include_formatting', 'bool', _('Include formatting tools in toolbar'), True),
			# T: option for plugin preferences
		('position', 'choice', _('Position in the window'), TOP, POSITIONS),
			# T: option for plugin preferences
		('style', 'choice', _('Toolbar style'), 'ICONS', STYLES),
			# T: option for plugin preferences
		('size', 'choice', _('Toolbar size'), 'SMALL_TOOLBAR', SIZES),
			# T: option for plugin preferences
	)


class ToolBarMainWindowExtension(PageViewExtension):

	def __init__(self, plugin, view):
		PageViewExtension.__init__(self, plugin, view)
		self.connectto(self.plugin.preferences, 'changed', self.on_preferences_changed)
		self.on_preferences_changed(self.plugin.preferences)
		self.connectto(self.obj.preferences, 'changed', self.on_pageview_preferences_changed)

	def on_preferences_changed(self, preferences):
		window = self.pageview.get_toplevel()
		with self.on_pageview_preferences_changed.blocked():
			self.obj.preferences['show_edit_bar'] = \
				not preferences['include_formatting'] # Force these prefs to always be in sync
		toolbar = window.setup_toolbar(
			show=preferences['show_toolbar'],
			position=preferences['position'],
		)
		toolbar.set_style(_get_style(self.plugin.preferences['style']))
		toolbar.set_icon_size(_get_size(self.plugin.preferences['size']))

	@SignalHandler
	def on_pageview_preferences_changed(self, preferences):
		# Force these prefs to always be in sync
		if self.plugin.preferences['include_formatting'] == preferences['show_edit_bar']:
			self.plugin.preferences['include_formatting'] = not preferences['show_edit_bar']

	def teardown(self):
		window = self.pageview.get_toplevel()
		window.setup_toolbar() # Restore default


class ToolbarTaskListWindowExtension(TaskListWindowExtension):
	# FIXME - should be extension of generic Window base class

	def __init__(self, plugin, window):
		TaskListWindowExtension.__init__(self, plugin, window)
		self.connectto(self.plugin.preferences, 'changed', self.on_preferences_changed)
		self.on_preferences_changed(self.plugin.preferences)

	def on_preferences_changed(self, preferences):
		if self.window._toolbar:
			self.window._toolbar.set_style(_get_style(self.plugin.preferences['style']))
			self.window._toolbar.set_icon_size(_get_size(self.plugin.preferences['size']))




# Copyright 2012 Yao-Po Wang <blue119@gmail.com>
# Copyright 2013 Jaap Karssenberg <jaap.karssenberg@gmail.com>


'''This module handles dumping reStructuredText with sphinx extensions'''

from zim.formats import *
from zim.formats.plain import Dumper as TextDumper

info = {
	'name': 'reST',
	'desc': 'reST (Octopress)',
	'mimetype': 'text/x-rst',
	'extension': 'rst',
		# No official file extension, but this is often used
	'native': False,
	'import': False,
	'export': True,
	'usebase': True,
}


class Dumper(TextDumper):

	BULLETS = {
		UNCHECKED_BOX: '- \u2610',
		XCHECKED_BOX: '- \u2612',
		CHECKED_BOX: '- \u2611',
		MIGRATED_BOX: '- \u25B7',
		TRANSMIGRATED_BOX: '- \u25C1',
		BULLET: '-',
	}

	TAGS = {
		EMPHASIS: ('*', '*'),
		STRONG: ('**', '**'),
		MARK: ('', ''), # TODO, no directly way to do this in rst
		STRIKE: ('', ''), # TODO, no directly way to do this in rst
		VERBATIM: ("``", "``"),
		TAG: ('', ''), # No additional annotation (apart from the visible @)
		SUBSCRIPT: ('\\ :sub:`', '`\\ '),
		SUPERSCRIPT: ('\\ :sup:`', '`\\ '),
	}
	# TODO tags other than :sub: and :sup: may also need surrounding whitespace, deal with this in post process (join) action ?
	# IDEM for blocks like images and objects, how to enforce empty lines and how to deal with inline images..

	HEADING_UNDERLINE = ['=', '-', '^', '"']

	def dump(self, tree):
		assert self.linker, 'rst dumper needs a linker object'
		return TextDumper.dump(self, tree)

	def dump_h(self, tag, attrib, strings):
		# Underlined headings
		level = int(attrib['level'])
		if level < 1:
			level = 1
		elif level > 4:
			level = 4
		char = self.HEADING_UNDERLINE[level - 1]
		heading = ''.join(strings)
		underline = char * len(heading.strip('\n'))
		return [heading, underline + '\n']

	def dump_pre(self, tag, attrib, strings):
		# prefix last line with "::\n\n"
		# indent with \t to get preformatted
		strings = self.prefix_lines('\t', strings)
		strings.insert(0, '::\n\n')
		return strings

	def dump_link(self, tag, attrib, strings=None):
		# Use inline url form, putting links at the end is more difficult
		assert 'href' in attrib, \
			'BUG: link misses href: %s "%s"' % (attrib, strings)
		href = self.linker.link(attrib['href'])
		text = ''.join(strings) or href
		return '`%s <%s>`_' % (text, href)

	def dump_img(self, tag, attrib, strings=None):
		src = self.linker.img(attrib['src'])
		text = '.. image:: %s\n' % src

		items = sorted(attrib.items())
		for k, v in items:
			if k == 'src' or k.startswith('_'):
				continue
			elif v: # skip None, "" and 0
				text += '   :%s: %s\n' % (k, v)

		return [text + '\n']

		# TODO use text for caption (with full recursion)
		# can be done using "figure" directive

	dump_object_fallback = dump_pre

	def dump_table(self, tag, attrib, strings):
		table = []  # result table

		aligns, _wraps = TableParser.get_options(attrib)
		rows = TableParser.convert_to_multiline_cells(strings)
		maxwidths = TableParser.width3dim(rows)
		rowsep = lambda y: TableParser.rowsep(maxwidths, x='+', y=y)
		rowline = lambda row: TableParser.rowline(row, maxwidths, aligns)

		# print table
		table.append(rowsep('-'))
		table += [rowline(line) for line in rows[0]]
		table.append(rowsep('='))
		for row in rows[1:]:
			table += [rowline(line) for line in row]
			table.append(rowsep('-'))

		return [line + "\n" for line in table]

	def dump_line(self, tag, attrib, strings=None):
		return '\n{}\n'.format('*' * 5)




# Copyright 2009-2018 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''Utilities to work with the clipboard for copy-pasting

Some functions defined here are also re-used for drag-and-drop
functionality, which works similar to the clipboard, but has a less
straight forward API.
'''

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf

import logging


import zim.datetimetz as datetime

from zim.fs import adapt_from_oldfs
from zim.parse.encode import url_encode, URL_ENCODE_READABLE
from zim.newfs import FilePath, LocalFile, LocalFolder
from zim.notebook import Path, HRef, set_parsetree_attributes_to_resolve_links, \
	replace_parsetree_links_and_copy_images
from zim.parse.links import is_url_re, link_type
from zim.formats import get_format, ParseTree, ParseTreeBuilder, \
	FORMATTEDTEXT, IMAGE, LINK, TAG, OBJECT
from zim.export.linker import StaticExportLinker


logger = logging.getLogger('zim.gui.clipboard')


# Targets have format (name, flags, id), flags need be 0 to allow
# paste to other widgets and to other application instances
PARSETREE_TARGET_ID = 1
PARSETREE_TARGET_NAME = 'text/x-zim-parsetree'
PARSETREE_TARGET = (PARSETREE_TARGET_NAME, 0, PARSETREE_TARGET_ID)

PAGELIST_TARGET_ID = 3
PAGELIST_TARGET_NAME = 'text/x-zim-page-list'
PAGELIST_TARGET = (PAGELIST_TARGET_NAME, 0, PAGELIST_TARGET_ID)

IMAGE_TARGET_ID = 6
IMAGE_TARGETS = tuple(
	(name, 0, IMAGE_TARGET_ID) for name in [
		'image/png', 'image/x-wmf', 'application/x-navi-animation',
		'image/bmp', 'image/x-bmp', 'image/x-MS-bmp', 'image/gif',
		'image/x-icns', 'image/x-icon', 'image/x-ico', 'image/x-win-bitmap',
		'image/vnd.microsoft.icon', 'application/ico', 'image/ico',
		'image/icon', 'text/ico', 'image/jpeg', 'image/x-portable-anymap',
		'image/x-portable-bitmap', 'image/x-portable-graymap',
		'image/x-portable-pixmap', 'image/x-quicktime', 'image/qtif',
		'image/svg+xml', 'image/svg', 'image/svg-xml',
		'image/vnd.adobe.svg+xml', 'text/xml-svg', 'image/svg+xml-compressed',
		'image/x-tga', 'image/tiff', 'image/x-xbitmap', 'image/x-xpixmap'
	] # TODO - check any win32 specific types to be added
)
#IMAGE_TARGETS = tuple(Gtk.target_list_add_image_targets(info=IMAGE_TARGET_ID))
IMAGE_TARGET_NAMES = tuple([target[0] for target in IMAGE_TARGETS])

# Add image format names as well, seen these being used by MS Office
for format in GdkPixbuf.Pixbuf.get_formats():
	if format.get_mime_types()[0] in IMAGE_TARGET_NAMES:
		name = format.get_name()
		for n in (name, name.upper()):
			IMAGE_TARGET_NAMES += (n,)
			IMAGE_TARGETS += ((n, 0, IMAGE_TARGET_ID),)

#~ print IMAGE_TARGETS
#~ print IMAGE_TARGET_NAMES

URI_TARGET_ID = 7
URI_TARGETS = (('text/uri-list', 0, URI_TARGET_ID),)
#URI_TARGETS = tuple(Gtk.target_list_add_uri_targets(info=URI_TARGET_ID))
	# According to docs we should provide list as arg to this function,
	# but seems docs are not correct
URI_TARGET_NAMES = tuple([target[0] for target in URI_TARGETS])

HTML_TARGET_ID = 8
HTML_TARGET_NAMES = ('text/html', 'HTML Format')
	# "HTML Format" is from MS Word
HTML_TARGETS = tuple([(name, 0, HTML_TARGET_ID) for name in HTML_TARGET_NAMES])

TEXT_TARGET_ID = 9
TEXT_TARGETS = tuple(
	(name, 0, TEXT_TARGET_ID) for name in [
		'UTF8_STRING', 'COMPOUND_TEXT', 'TEXT', 'STRING',
		'text/plain;charset=utf-8', 'text/plain'
	]
)
#TEXT_TARGETS = tuple(Gtk.target_list_add_text_targets(info=TEXT_TARGET_ID))
	# According to docs we should provide list as arg to this function,
	# but seems docs are not correct
TEXT_TARGET_NAMES = tuple([target[0] for target in TEXT_TARGETS])

# All targets that we can convert to a parsetree, in order of choice
PARSETREE_ACCEPT_TARGETS = (PARSETREE_TARGET, PAGELIST_TARGET) \
	+ IMAGE_TARGETS + URI_TARGETS + TEXT_TARGETS
PARSETREE_ACCEPT_TARGET_NAMES = tuple([target[0] for target in PARSETREE_ACCEPT_TARGETS])
#~ print('ACCEPT', PARSETREE_ACCEPT_TARGET_NAMES)


_internal_selection_data = None # HACK used for workaround for issue #390


# Mimetype text/uri-list is used for drag n drop of URLs
# it is plain text encoded list of urls, separated by \r\n
# Since not all apps follow the standard exactly, do allow for
# just \n separated lists.
#
# Zim page links are also encoded in this format, but no url encoding
# is applied and no scheme is used. The target mimetype can be used to
# distinguish between proper uri lists and lists of page names.

def pack_urilist(links):
	text = ''
	for link in links:
		if is_url_re.match(link):
			link = url_encode(link, mode=URL_ENCODE_READABLE) # just to be sure
		text += '%s\r\n' % link
	return text.encode()


def unpack_urilist(text):
	# FIXME be tolerant here for file://path/to/file uris here
	text = text.strip(b'\x00').decode() # Found trailing NULL character on windows
	lines = text.splitlines() # takes care of \r\n
	return [line for line in lines if line and not line.isspace()]
		# Just to be sure we also skip empty or whitespace lines

# TODO: Probably the serialize formats can replace custom copy/paste
# handlers in TextView and TextBuffer as well


def textbuffer_register_serialize_formats(buffer, notebook, page):
	buffer.register_serialize_format('text/x-zim-parsetree', serialize_parse_tree)
	buffer.register_deserialize_format('text/x-zim-parsetree', deserialize_parse_tree, (notebook, page))
	for name in (PAGELIST_TARGET_NAME,) + URI_TARGET_NAMES:
		buffer.register_deserialize_format(name, deserialize_urilist, (notebook, page))
	for name in IMAGE_TARGET_NAMES: # FIXME, should we limit the list ?
		buffer.register_deserialize_format(name, deserialize_image, (name, notebook, page))

def serialize_parse_tree(register_buf, content_buf, start, end, user_data):
	tree = content_buf.get_parsetree((start, end))
	xml = tree.tostring()
	return xml

def deserialize_parse_tree(register_buf, content_buf, iter, data, length, create_tags, user_data):
	notebook, path = user_data
	tree = ParseTree().fromstring(data)
	content_buf.insert_parsetree(iter, tree, interactive=True)
	return True

def deserialize_urilist(register_buf, content_buf, iter, data, length, create_tags, user_data):
	notebook, path = user_data
	links = unpack_urilist(data)
	tree = _link_tree(links, notebook, path)
	content_buf.insert_parsetree(iter, tree, interactive=True)
	return True


def _get_paste_image_file(dir, notebook, extension):
	tmpl = notebook.config['Notebook'].get('paste_image_template', 'pasted_image')
	name = datetime.strftime(tmpl, datetime.now()).rstrip('.') + '.' + extension
	return dir.new_file(name)


def deserialize_image(register_buf, content_buf, iter, data, length, create_tags, user_data):
	# Implementation note: we follow gtk_selection_get_pixbuf() in usage of
	# Gtk.PixbufLoader to capture clipboard data in a pixbuf object.
	# We could skip this, but it allows for on-the-fly conversion of the data
	# type.
	mimetype, notebook, path = user_data

	# capture image
	loader = GdkPixbuf.PixbufLoader()
	loader.write(data)
	loader.close()
	pixbuf = loader.get_pixbuf()

	# save it as an attachment
	dir = notebook.get_attachments_dir(path)
	if not dir.exists():
		logger.debug("Creating attachment dir: %s", dir)
		dir.touch()

	format, extension = _get_image_info(mimetype)
	if format is None or format == 'bmp':
		# default to png format
		# special casing bmp since many window apps use it internally
		# but is quite large to store, so compress by using png
		format, extension = 'png', 'png'

	file = _get_paste_image_file(dir, notebook, extension)
	logger.debug("Saving image from clipboard to %s", file)
	file.touch() # notify version control
	pixbuf.savev(file.path, format, [], [])

	# and insert it in the page
	links = [file.uri]
	tree = _link_tree(links, notebook, path)
	content_buf.insert_parsetree(iter, tree, interactive=True)
	return True


def parsetree_from_selectiondata(selectiondata, notebook, path=None, text_format='plain'):
	'''Function to get a parsetree based on the selectiondata contents
	if at all possible. Used by both copy-paste and drag-and-drop
	methods.

	The 'notebook' and optional 'path' arguments are used to format
	links relative to the page which is the target for the pasting or
	drop operation.

	NOTE: this method can have a side affect where files are placed in the
	attachment folder of the page.

		- For image data, a new image file is created
		- For parsetree data containing inline images, the image files are
	  	  copied to the attachment folder

	@param selectiondata: a C{Gtk.SelectionData} object
	@param notebook: a L{Notebook} object
	@param path: a L{Path} object
	@param text_format: format to parse pasted text, as a special case

		- "verbatim" will wrap content in VERBARIM_BLOCK or VERBATIM element based on the content
		- "verbatim-pre" will wrap the content in a VERBATIM_BLOCK element and
		- "verbatim-code" will wrap the content in a VERBATIM element

	@returns: a L{ParseTree} or C{None}
	'''
	targetname = selectiondata.get_target().name()
	if targetname == PARSETREE_TARGET_NAME:
		parsetree = ParseTree().fromstring(selectiondata.get_data())
		return replace_parsetree_links_and_copy_images(parsetree, notebook, path)
	elif targetname == PAGELIST_TARGET_NAME or targetname in URI_TARGET_NAMES:
		links = selectiondata.get_uris()
		return _link_tree(links, notebook, path)
	elif targetname in TEXT_TARGET_NAMES:
		# plain text parser should highlight urls etc.
		# FIXME some apps drop text/uri-list as a text/plain mimetype
		# try to catch this situation by a check here
		text = selectiondata.get_text()
		if text:
			if text_format in ('verbatim', 'verbatim-pre', 'verbatim-code'):
				if text_format == 'verbatim':
					tag_name = 'pre' if '\n' in text else 'code'
				else:
					tag_name = text_format[9:]
				builder = ParseTreeBuilder()
				builder.start('zim-tree', {})
				builder.start(tag_name, {})
				builder.text(text)
				builder.end(tag_name)
				builder.end('zim-tree')
				return builder.get_parsetree()
			else:
				return get_format(text_format).Parser().parse(text)
		else:
			return None
	elif targetname in IMAGE_TARGET_NAMES:
		# save image
		pixbuf = selectiondata.get_pixbuf()
		if not pixbuf:
			return None

		dir = notebook.get_attachments_dir(path)
		assert isinstance(dir, LocalFolder) or hasattr(dir, '_folder') and isinstance(dir._folder, LocalFolder)
			# XXX: assert we have local path  - HACK to deal with FilesAttachmentFolder
		if not dir.exists():
			logger.debug("Creating attachment dir: %s", dir)
			dir.touch()

		format, extension = _get_image_info(targetname)
		if format is None or format == 'bmp':
			# default to png format
			# special casing bmp since many window apps use it internally
			# but is quite large to store, so compress by using png
			format, extension = 'png', 'png'

		file = _get_paste_image_file(dir, notebook, extension)
		logger.debug("Saving image from clipboard to %s", file)
		file.touch() # notify version control
		pixbuf.savev(file.path, format, [], [])

		links = [file.uri]
		return _link_tree(links, notebook, path)
	else:
		return None


def _link_tree(links, notebook, path):
	# Convert a list of links (of any type) into a parsetree
	# To be used on the recieving end of a copy-paste, arguments are the target notebook and path
	#~ print('LINKS: ', links)
	#~ print('NOTEBOOK and PATH:', notebook, path)
	builder = ParseTreeBuilder()
	builder.start(FORMATTEDTEXT)
	for i in range(len(links)):
		if i > 0:
			builder.text(' ')

		link = links[i]
		type = link_type(link)
		isimage = False
		if type == 'interwiki':
			prefix = notebook.interwiki + '?'
			if link.startswith(prefix):
				link = link[len(prefix):]
				type = link_type(link)
		elif type == 'file':
			try:
				file = LocalFile(link)
			except ValueError:
				pass
			else:
				isimage = file.isimage()

		logger.debug('Pasting link: %s (type: %s, isimage: %s)', link, type, isimage)

		if isimage:
			src = notebook.relative_filepath(file, path) or file.uri
			builder.append(IMAGE, {'src': src})
		elif link.startswith('@'):
			# FIXME - is this ever used ??
			builder.append(TAG, {'name': links[i][1:]}, links[i])
		else:
			name = None
			if type == 'page':
				anchor = None
				if '#' in link:
					link, anchor = link.split('#', 1)
				target = Path(Path.makeValidPageName(link)) # Assume links are always absolute
				href = notebook.pages.create_link(path, target)
				href.anchor = anchor
				link = href.to_wiki_link()
				if notebook.config['Notebook']['short_links']:
					name = href.short_name()
			elif type == 'file':
				try:
					file = LocalFile(link) # Assume links are always URIs
					link = notebook.relative_filepath(file, path) or file.uri
				except:
					pass

			builder.append(LINK, {'href': link}, name or link)

	builder.end(FORMATTEDTEXT)
	tree = builder.get_parsetree()
	tree.decode_urls()
	return tree


def _get_image_info(targetname):
	# Target name for images is supposed to be mimetype, we check
	# in available pixbuf writers for this type and return the
	# format name and file extension
	for format in GdkPixbuf.Pixbuf.get_formats():
		name = format.get_name()
		mimetypes = format.get_mime_types()
		if targetname in (name, name.upper()) or targetname in mimetypes:
			if format.is_writable():
				return name, format.get_extensions()[0]
			else:
				return None, None
	else:
		return None, None


class MockSelectionData(object):
	'''Adapter to allow usage of C{ClipboardData} as input for
	C{parsetree_from_selectiondata()}
	'''

	def __init__(self, target, clipboard_data):
		self.target = target
		self.data = clipboard_data

	def get_target(self):
		target_name = self.target[0]
		return Gdk.Atom.intern(target_name, False)

	def get_data(self):
		target_id = self.target[-1]
		return self.data.get_data_as(target_id)

	def get_text(self):
		return self.data.get_data_as(TEXT_TARGET_ID)

	def get_uris(self):
		if self.target in URI_TARGETS:
			return self.data.get_data_as(URI_TARGET_ID)
		else:
			return unpack_urilist(self.get_data())

	def get_pixbuf(self):
		raise NotImplementedError


class ClipboardData(object):
	'''Wrapper for data that can be set on the clipboard and pasted
	multiple formats
	'''

	targets = ()

	def get_data_as(self, targetid):
		'''Return data in the requested target format
		@param targetid: the target id
		@implementation: must be implemented by sub-classes
		'''
		raise NotImplementedError


class UriData(ClipboardData):

	targets = URI_TARGETS + TEXT_TARGETS

	def __init__(self, *obj):
		uris = []
		text = []
		for o in obj:
			o = adapt_from_oldfs(o)
			if isinstance(o, FilePath):
				uris.append(o.uri)
				text.append(o.userpath)
			else:
				uri = o if isinstance(o, str) else o.uri
				uris.append(uri)
				text.append(uri)
		self.uris = tuple(uris)
		self.text = ' '.join(text)

	def get_data_as(self, targetid):
		if targetid == URI_TARGET_ID:
			return self.uris
		else:
			return self.text


class InterWikiLinkData(UriData):

	targets = (PARSETREE_TARGET,) + UriData.targets

	def __init__(self, href, url):
		self.interwiki_href = href
		self.interwiki_url = url
		if self.interwiki_url is not None:
			UriData.__init__(self, url)
		else:
			pass # interwiki may be undefined, resulting in url being None

	def get_data_as(self, targetid):
		if targetid == PARSETREE_TARGET_ID:
			builder = ParseTreeBuilder()
			builder.start(FORMATTEDTEXT)
			builder.append(LINK, {'href': self.interwiki_href}, self.interwiki_href)
			builder.end(FORMATTEDTEXT)
			parsetree = builder.get_parsetree()
			parsetree._set_root_attrib('notebook', '-') # force resolve on paste
			parsetree._set_root_attrib('page', '-')
			return parsetree.tostring()
		elif self.interwiki_url is not None:
			return UriData.get_data_as(self, targetid)
		else:
			return self.interwiki_href


class ParseTreeData(ClipboardData):

	targets = (PARSETREE_TARGET,) + HTML_TARGETS + TEXT_TARGETS

	def __init__(self, notebook, path, parsetree, format):
		self.notebook = notebook # FIXME - should not need to keep this reference - resolve parsetree immediatly
		self.path = path
		self.parsetree = parsetree
		self.format = format

	def get_data_as(self, targetid):
		if targetid == PARSETREE_TARGET_ID:
			newtree = set_parsetree_attributes_to_resolve_links(self.parsetree, self.notebook, self.path)
			return newtree.tostring()
		elif targetid == HTML_TARGET_ID:
			dumper = get_format('html').Dumper(
				linker=StaticExportLinker(self.notebook, source=self.path))
			html = ''.join(dumper.dump(self.parsetree))
			return wrap_html(html, self.format)
		elif targetid == TEXT_TARGET_ID:
			if self.format in ('wiki', 'plain'):
				dumper = get_format(self.format).Dumper()
			else:
				dumper = get_format(self.format).Dumper(
					linker=StaticExportLinker(self.notebook, source=self.path))

			return ''.join(dumper.dump(self.parsetree))
		else:
			raise ValueError('Unknown target id %i' % targetid)


class PageLinkData(ClipboardData):

	targets = (PARSETREE_TARGET, PAGELIST_TARGET) + TEXT_TARGETS

	def __init__(self, notebook, path, anchor=None, text=None):
		assert isinstance(path, Path)
		self.notebook = notebook
		self.path = path
		self.anchor = anchor
		self.text = text

	def get_data_as(self, targetid):
		if targetid == PAGELIST_TARGET_ID:
			link = "%s?%s" % (self.notebook.interwiki, self.path.name)
			if self.anchor:
				link += "#" + self.anchor
			return pack_urilist((link,))
		elif targetid == TEXT_TARGET_ID:
			link = self.path.name
			if self.anchor:
				link += "#" + self.anchor
			return link
		elif targetid == PARSETREE_TARGET_ID:
			link = self.path.name
			if self.anchor:
				link += "#" + self.anchor
			if self.text:
				text = self.text
			elif self.notebook.config['Notebook']['short_links']:
				href = HRef.new_from_wiki_link(link)
				text = href.short_name()
			else:
				text = link

			# Same logic as set_parsetree_attributes_to_resolve_links() but no need to resolve again
			builder = ParseTreeBuilder()
			builder.start(FORMATTEDTEXT)
			builder.append(LINK, {'href': link, '_href': self.path.name}, text)
			builder.end(FORMATTEDTEXT)
			parsetree = builder.get_parsetree()
			parsetree._set_root_attrib('notebook', self.notebook.interwiki)
			parsetree._set_root_attrib('page', '-') # force resolve on paste also on same page
			return parsetree.tostring()
		else:
			raise ValueError('Unknown target id %i' % targetid)


class ClipboardManager(object):
	'''Wrapper for C{Gtk.Clipboard}, supporting specific data formats'''

	def __init__(self, name):
		'''Constructor
		@param name: clipboard name, can be either "CLIPBOARD" or "PRIMARY",
		see C{Gtk.Clipboard} for details.
		'''
		assert name in ('CLIPBOARD', 'PRIMARY')
		atom = Gdk.SELECTION_CLIPBOARD if name == 'CLIPBOARD' else Gdk.SELECTION_PRIMARY
		self.clipboard = Gtk.Clipboard.get(atom)
		self.data = None

	def clear(self):
		if self.data is not None:
			self.clipboard.clear()
		self.data = None

	def set_clipboard_data(self, clipboard_data):
		'''Set an item on the clipboard
		@param clipboard_data: a L{ClipboardData} object
		'''
		self.data = clipboard_data
		#self.clipboard.set_with_data(self.data.targets, self._get, self._clear) \
		#	or logger.warn('Failed to set data on clipboard')

		### set_with_data() workaround
		assert TEXT_TARGETS[0] in self.data.targets
		text = self.data.get_data_as(TEXT_TARGET_ID)
		self.clipboard.set_text(text, -1)
		self.data._workaround_text = text
		###

	def _get(self, clipboard, selectiondata, targetid):
		logger.debug(
			"Clipboard requests data as '%s', we have %r",
			selectiondata.get_target().name(),
			self.data
		)
		data = self.data.get_data_as(targetid)
		if targetid == TEXT_TARGET_ID:
			selectiondata.set_text(data, -1)
		elif targetid == URI_TARGET_ID:
			selectiondata.set_uris(data)
		else:
			selectiondata.set(PARSETREE_TARGET_NAME, 8, data)

	def _clear(self):
		self.data = None

	def set_text(self, text):
		'''Set text to the clipboard
		@param text: text to set on the clipboard
		@note: DO NOT USE THIS METHOD if you can use L{set_parsetree()}
		instead
		'''
		self.clipboard.set_text(text, -1)

	def get_text(self):
		'''Get text from the clipboard.
		@returns: (unicode) text or C{None}
		@note: DO NOT USE THIS METHOD if you can use L{get_parsetree()}
		instead
		'''
		return self.clipboard.wait_for_text()

	def set_parsetree(self, notebook, path, parsetree, format='plain'):
		'''Copy a parsetree to the clipboard. The parsetree can be pasted by
		the user either as formatted text within zim or as plain text outside
		zim. The tree can be the full tree for 'page', but also a selection.

		@param notebook: the L{Notebook} object
		@param path: the L{Path} object - used to resolve links etc.
		@param parsetree: the actual L{ParseTree} to be set on the clipboard
		@keyword format: the format to use for pasting text, e.g. 'wiki' or 'plain'
		'''
		self.set_clipboard_data(
			ParseTreeData(notebook, path, parsetree, format) )

	def get_parsetree(self, notebook=None, path=None, text_format='plain'):
		'''Get a parsetree from the clipboard.

		Can handle various data types and convert them to L{ParseTree}
		objects. So they can be pasted directly in a text buffer.

		The 'notebook' and optional 'path' arguments are used to format
		links relative to the page which is the target for the pasting or
		drop operation. Otherwise absolute links will be used.

		@param notebook: a L{Notebook} object
		@param path: a L{Path} object
		@param text_format: format to parse pasted text

		@returns: a L{ParseTree} or C{None}
		'''
		(has_data, atoms) = self.clipboard.wait_for_targets()
		logger.debug('Targets available for paste: %s, we want parsetree', atoms)

		if not has_data:
			return None

		### set_with_data() workaround
		if atoms and self.data and hasattr(self.data, '_workaround_text') \
			and any(a.name() in TEXT_TARGET_NAMES for a in atoms):
				text = self.clipboard.wait_for_text()
				if text == self.data._workaround_text:
					targets = sorted(
						filter(lambda t: t in PARSETREE_ACCEPT_TARGETS, self.data.targets),
						key=lambda t: PARSETREE_ACCEPT_TARGETS.index(t)
					)
					assert len(targets) > 0
					logger.debug('Requesting data for %s -- using set_with_data() workaround', targets[0])
					selectiondata = MockSelectionData(targets[0], self.data)
					return parsetree_from_selectiondata(selectiondata, notebook, path, text_format)
		###

		atoms = sorted(
					filter(lambda a: a and a.name() in PARSETREE_ACCEPT_TARGET_NAMES, atoms),
					key=lambda a: PARSETREE_ACCEPT_TARGET_NAMES.index(a.name())
				)
				# NOTE: "atoms" can contain None values, see issue #774
		if atoms:
			atom = atoms[0]  # TODO why choose 1st index?
			logger.debug('Requesting data for %s', atom)
			selectiondata = self.clipboard.wait_for_contents(atom)
			if selectiondata:
				return parsetree_from_selectiondata(selectiondata, notebook, path, text_format)
			else:
				logger.warning('Did not get requested data from clipboard')
				return None
		else:
			logger.warning('Could not paste - no compatible data types on clipboard')
			return None

	def set_pagelink(self, notebook, path, anchor=None, text=None):
		'''Copy a page name to the clipboard. The page name can be pasted by the
		user either as a link within zim or as text outside zim.
		@param notebook: a L{Notebook} object
		@param path: a L{Path} object
		'''
		logger.debug("set_pagelink %r %r anchor=%s text=%s", notebook, path, anchor, text)
		self.set_clipboard_data(PageLinkData(notebook, path, anchor, text))

	def set_interwikilink(self, href, url):
		'''Copy an interwiki link to the clipboard
		@param href: the link as shown in zim, e.g. "wp?foobar"
		@param url: the expanded url for this interwiki link, e.g.
		"http://en.wikipedia.org/wiki/foobar"
		'''
		self.set_clipboard_data(InterWikiLinkData(href, url))

	def set_uri(self, *uris):
		'''Copy an uri to the clipboard
		@param uri: an uri as string, or an object with an attribute C{uri}
		'''
		self.set_clipboard_data(UriData(*uris))


Clipboard = ClipboardManager("CLIPBOARD") #: Singleton object for the default clipboard


SelectionClipboard = ClipboardManager("PRIMARY") #: Singleton object for the selection clipboard (unix)




########### Code to deal with HTML formatting on windows ############

HTML_HEAD = '''\
<meta http-equiv="Content-Type" content="text/html;charset=utf-8">
<meta name="Description" content="Copy-Paste Buffer">
<meta name="Generator" content="Zim">'''

def wrap_html(html, target):
	'''Function to wrap html with appropriate headers based on target type'''
	if target == 'HTML Format':
		return Win32HtmlFormat.encode(html, head=HTML_HEAD)
	else:
		return '''\
<html>
<head>
%s
</head>
<body>
%s
</body>
</html>
''' % (HTML_HEAD, html)


class Win32HtmlFormat:
	'''This class adds support for Windows "HTML Format" clipboard content type

	Code is based on example code from
	U{http://code.activestate.com/recipes/474121/}

	written by Phillip Piper (jppx1[at]bigfoot.com)

	Also see specification at:
	U{http://msdn.microsoft.com/library/default.asp?url=/library/en-us/winui/winui/windowsuserinterface/dataexchange/clipboard/htmlclipboardformat.asp}
	'''

	MARKER_BLOCK_OUTPUT = \
		"Version:1.0\r\n" \
		"StartHTML:%09d\r\n" \
		"EndHTML:%09d\r\n" \
		"StartFragment:%09d\r\n" \
		"EndFragment:%09d\r\n" \
		"StartSelection:%09d\r\n" \
		"EndSelection:%09d\r\n" \
		"SourceURL:%s\r\n"

	#~ MARKER_BLOCK_EX = \
		#~ "Version:(\S+)\s+" \
		#~ "StartHTML:(\d+)\s+" \
		#~ "EndHTML:(\d+)\s+" \
		#~ "StartFragment:(\d+)\s+" \
		#~ "EndFragment:(\d+)\s+" \
		#~ "StartSelection:(\d+)\s+" \
		#~ "EndSelection:(\d+)\s+" \
		#~ "SourceURL:(\S+)"
	#~ MARKER_BLOCK_EX_RE = re.compile(MARKER_BLOCK_EX)

	#~ MARKER_BLOCK = \
		#~ "Version:(\S+)\s+" \
		#~ "StartHTML:(\d+)\s+" \
		#~ "EndHTML:(\d+)\s+" \
		#~ "StartFragment:(\d+)\s+" \
		#~ "EndFragment:(\d+)\s+" \
		#~ "SourceURL:(\S+)"
	#~ MARKER_BLOCK_RE = re.compile(MARKER_BLOCK)

	DEFAULT_HTML_BODY = \
		"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">" \
		"<HTML><HEAD>%s</HEAD><BODY><!--StartFragment-->%s<!--EndFragment--></BODY></HTML>"

	@classmethod
	def encode(klass, fragment, selection=None, head=None, source=None):
		if selection is None:
			selection = fragment
		if source is None:
			source = "zim://copy-pase-buffer"
		if head is None:
			head = ''

		html = klass.DEFAULT_HTML_BODY % (head, fragment)
		fragmentStart = html.index(fragment)
		fragmentEnd = fragmentStart + len(fragment)
		selectionStart = html.index(selection)
		selectionEnd = selectionStart + len(selection)

		# How long is the prefix going to be?
		dummyPrefix = klass.MARKER_BLOCK_OUTPUT % (0, 0, 0, 0, 0, 0, source)
		lenPrefix = len(dummyPrefix)

		prefix = klass.MARKER_BLOCK_OUTPUT % (
			lenPrefix, len(html) + lenPrefix,
			fragmentStart + lenPrefix, fragmentEnd + lenPrefix,
			selectionStart + lenPrefix, selectionEnd + lenPrefix,
			source
		)
		return prefix + html

	#~ @classmethod
	#~ def decode(self, data):
		#~ """
		#~ Decode the given string to figure out the details of the HTML that's on the string
		#~ """
		#~ # Try the extended format first (which has an explicit selection)
		#~ matches = self.MARKER_BLOCK_EX_RE.match(src)
		#~ if matches:
			#~ self.prefix = matches.group(0)
			#~ self.htmlClipboardVersion = matches.group(1)
			#~ self.html = src[int(matches.group(2)):int(matches.group(3))]
			#~ self.fragment = src[int(matches.group(4)):int(matches.group(5))]
			#~ self.selection = src[int(matches.group(6)):int(matches.group(7))]
			#~ self.source = matches.group(8)
		#~ else:
			#~ # Failing that, try the version without a selection
			#~ matches = self.MARKER_BLOCK_RE.match(src)
			#~ if matches:
				#~ self.prefix = matches.group(0)
				#~ self.htmlClipboardVersion = matches.group(1)
				#~ self.html = src[int(matches.group(2)):int(matches.group(3))]
				#~ self.fragment = src[int(matches.group(4)):int(matches.group(5))]
				#~ self.source = matches.group(6)
				#~ self.selection = self.fragment




# Copyright 2013 Jaap Karssenberg <jaap.karssenberg@gmail.com>

from gi.repository import Gtk
from gi.repository import GObject
import logging

import zim.datetimetz as datetime

from zim.notebook import Path
from zim.gui.widgets import Dialog, BrowserTreeView, ScrolledWindow


logger = logging.getLogger('zim.gui.dialogs')


class RecentChangesDialog(Dialog):

	def __init__(self, widget, notebook, navigation):
		Dialog.__init__(self, widget, _('Recent Changes'), # T: Dialog title
			buttons=Gtk.ButtonsType.CLOSE,
			defaultwindowsize=(400, 300)
		)
		self.notebook = notebook
		self.navigation = navigation

		self.treeview = RecentChangesTreeView()
		self.vbox.pack_start(ScrolledWindow(self.treeview), True, True, 0)
		self.treeview.connect('row-activated', self.on_row_activated)

		self.update()
		self.notebook.connect_after('stored-page', lambda *a: self.update())

	def update(self):
		model = self.treeview.get_model()
		if model is not None:
			model.clear()
			for rec in self.notebook.pages.list_recent_changes(limit=50):
				model.append((rec.name, rec.mtime))
		# else already destroyed ?

	def on_row_activated(self, view, path, col):
		page = Path(view.get_model()[path][view.NAME_COL])
		self.navigation.open_page(page)


class RecentChangesTreeView(BrowserTreeView):

	NAME_COL = 0
	MODIFIED_COL = 1

	def __init__(self):
		model = Gtk.ListStore(str, float)
			# NAME_COL, MODIFIED_COL
		BrowserTreeView.__init__(self, model)

		cell_renderer = Gtk.CellRendererText()

		column = Gtk.TreeViewColumn(_('Page'), cell_renderer, text=self.NAME_COL) # T: Column header
		column.set_sort_column_id(self.NAME_COL)
		column.set_expand(True)
		self.append_column(column)

		today = datetime.date.today()
		yesterday = today - datetime.timedelta(days=1)
		def render_date(col, cell, model, i, data):
			mtime = model.get_value(i, self.MODIFIED_COL)
			if mtime:
				dt = datetime.datetime.fromtimestamp(mtime)
				date = dt.date()
				if date == today:
					text = _('Today') + datetime.strftime(' %H:%M', dt)
					# T: label for modified time
				elif date == yesterday:
					text = _('Yesterday') + datetime.strftime(' %H:%M', dt)
					# T: label for modified time
				elif date.year == today.year:
					text = datetime.strftime('%a %d %b %H:%M', dt) # TODO allow config for format ?
				else:
					text = datetime.strftime('%a %d %b %Y %H:%M', dt) # TODO allow config for format ?
			else:
				text = ''

			cell.set_property('text', text)

		cell_renderer = Gtk.CellRendererText()
		#cell_renderer.set_property('font', 'mono')
		column = Gtk.TreeViewColumn(_('Last Modified'), cell_renderer) # T: Column header
		column.set_cell_data_func(cell_renderer, render_date)
		column.set_sort_column_id(self.MODIFIED_COL)
		self.append_column(column)




# Copyright 2009-2012 Jaap Karssenberg <jaap.karssenberg@gmail.com>
# Copyright 2010,2011 John Drinkwater <john@nextraweb.com>
# Copyright 2012 Damien Accorsi <damien.accorsi@free.fr>



import os
import logging

from zim.plugins.versioncontrol import VCSApplicationBase
from zim.applications import Application


logger = logging.getLogger('zim.vcs.git')


class GITApplicationBackend(VCSApplicationBase):

	@classmethod
	def build_bin_application_instance(cls):
		return Application(('git',))

	def build_revision_arguments(self, versions, is_for_diff=False):
		"""Build a list including required string/int for running an VCS command
		# Accepts: None, int, string, (int,), (int, int)
		# Always returns a list
		# versions content:
		  - None: return an empty list
		  - int ou string: return ['-r', int]
		  - tuple or list: return ['-r', '%i..%i']

		It's all based on the fact that defining revision with current VCS is:
		-r revision
		-r rev1..rev2
		"""
		if is_for_diff:
			if len(versions) == 2:
				versions.reverse()
				return ['..'.join(versions)]
			elif len(versions) == 1:
				return [versions[0] + '^']
			else:
				return []
		else:
			if isinstance(versions, (tuple, list)):
				assert 1 <= len(versions) <= 2
				if len(versions) == 2:
					return list(map(str, versions))
				else:
					versions = versions[0]

			if not versions is None:
				version = str(versions)
				return [version]
			else:
				return []

	def path_arg(self, path):
		# make sure "/" is used as path seperator as git under msys fails otherwise
		if isinstance(path, str):
			return path.replace('\\', '/')
		else:
			return path.relpath(self.root).replace('\\', '/')

	########
	#
	# NOW ARE ALL REVISION CONTROL SYSTEM SHORTCUTS

	def add(self, path=None):
		"""
		Runs: git add {{PATH}}
		"""
		if path is None:
			return self.run(['add', self.notebook_dir])
		else:
			return self.run(['add', self.path_arg(path)])


	def annotate(self, file, version=None):
		"""FIXME Document
		return
		0: line1
		2: line1
		...
		"""
		revision_args = self.build_revision_arguments(version)
		return self.pipe(['blame', '-s', self.path_arg(file)] + revision_args)

	def cat(self, path, version):
		"""
		Runs: git cat {{PATH}} {{REV_ARGS}}
		"""
		revision_args = self.build_revision_arguments(version)
		return self.pipe(['show', ''.join([''.join(revision_args), ':', self.path_arg(path)])])

	def commit(self, path, msg):
		"""
		Runs: git commit -m {{MSG}} {{PATH}}
		"""
		if self.is_modified():
			params = ['commit']
			if msg != '' and msg is not None:
				params.append('-m')
				params.append(msg)
			if path != '' and path is not None:
				params.append('--')
				params.append(self.path_arg(path))
			return self.run(params)

	def diff(self, versions=None, file=None):
		"""
		Runs:
			git diff --no-ext-diff {{REVISION_ARGS}}
		or
			git diff --no-ext-diff {{REVISION_ARGS}} -- {{PATH}}
		"""
		revision_args = self.build_revision_arguments(versions)
		revision_args = self.build_revision_arguments(revision_args, is_for_diff=True)
		if file is None:
			return self.pipe(['diff', '--no-ext-diff'] + revision_args)
		else:
			return self.pipe(['diff', '--no-ext-diff'] + revision_args + ['--', self.path_arg(file)])

	def ignore(self, file_to_ignore_regexp):
		"""
		Build a .gitignore file including the file_to_ignore_content
		"""
		#TODO: append the rule instead of overwrite the full content
		self.root.file('.gitignore').write(file_to_ignore_regexp)

	def init_repo(self):
		self.init()
		# Configure git to not quote non-ascii filenames. By default,
		# they are quoted as octal escapes, making it complicated to
		# deal with page names in non-Latin scripts.
		self.run(['config', 'core.quotepath', 'off'])
		self.ignore(".zim/\n")
		self.add('.') # add all existing files

	def repo_exists(self):
		return self.root.folder('.git').exists() or self.root.file('.git').exists()

	def init(self):
		"""
		Runs: git init
		"""
		return self.run(['init'])

	def is_modified(self):
		"""Returns true if the repo is not up-to-date, or False
		@returns: True if the repo is not up-to-date, or False
		"""
		# If status return an empty answer, this means the local repo is up-to-date
		status = ''.join(self.pipe(['status', '--porcelain']))
		return bool(status.strip())

	def log(self, path=None):
		"""
		Runs:
			git log --date=iso --follow {{PATH}}
		or
			git log --date=iso
		"""
		if path:
			return self.pipe(['log', '--date=iso', '--follow', self.path_arg(path)])
		else:
			return self.pipe(['log', '--date=iso'])

	def log_to_revision_list(self, log_op_output):
		versions = []
		(rev, date, user, msg) = (None, None, None, None)
		seenmsg = False
		# seenmsg allow to get the complete commit message which is presented like this:
		#
		# [...]
		# description:
		# here is the
		# commit message
		# the end of it may be detected
		# because of the apparition of a line
		# starting by "changeset:"
		#
		# FIXME: there is a bug which will stop parsing if a blank line is included
		# in the commit message
		for line in log_op_output:
			if line.startswith('commit '):
				if not rev is None:
					versions.append((rev, date, user, msg))
				(rev, date, user, msg) = (None, None, None, None)
				seenmsg = False
				rev = line[7:].strip()
			elif line.startswith('Author: '):
				user = line[7:].strip()
			elif line.startswith('Date: '):
				date = line[7:].strip()
				seenmsg = True
				msg = ''
			elif seenmsg and line.startswith(' '):
				msg += line[4:]

		if not rev is None:
			versions.append((rev, date, user, msg))

		versions.reverse()
		return versions


	def move(self, oldpath, newpath):
		return self.add(newpath) # move already happened

	def remove(self, path):
		pass # delete already happened

	def revert(self, path=None, version=None):
		"""
		Runs:
			hg revert {{PATH}} {{REV_ARGS}}
			is equivalent to
			git checkout {{REV_ARGS}} -- {{PATH}}

		or
			hg revert --no-backup --all {{REV_ARGS}}
			is equivalent to
			git reset --hard HEAD
		"""
		revision_params = self.build_revision_arguments(version)
		if path:
			self.run(['checkout'] + revision_params + ['--', self.path_arg(path)])
		else:
			self.run(['reset', '--hard', 'HEAD'])

	def stage(self):
		self.add()

	def status(self, porcelain=False):
		"""
		Runs: git status
		@param porcelain: see --porcelain in git documentation, used for testing
		"""
		if porcelain:
			return self.pipe(['status', '--porcelain'])
		else:
			return self.pipe(['status'])




# Copyright 2009-2015 Jaap Karssenberg <jaap.karssenberg@gmail.com>

# Tests: search gui.TestDialogs.testSearchDialog

from gi.repository import Gtk
from gi.repository import GObject
import logging

from zim.notebook import Path
from zim.gui.widgets import Dialog, BrowserTreeView, InputEntry, ErrorDialog, ScrolledWindow, StatusPage
from zim.gui.pageview.find import FIND_REGEX

from zim.search import *

logger = logging.getLogger('zim.gui.searchdialog')

HELP_TEXT = _(
	'For advanced search you can use operators like\n'
	'AND, OR and NOT. See the help page for more details.'
) # T: help text for the search dialog

class SearchDialog(Dialog):

	READY = 0
	SEARCHING = 1
	DONE = 2
	CANCELLED = 3

	def __init__(self, widget, notebook, page, navigation):
		Dialog.__init__(self, widget, _('Search'), # T: Dialog title
			buttons=Gtk.ButtonsType.CLOSE, help='Help:Searching',
			defaultwindowsize=(400, 300)
		)
		self.page = page

		hbox = Gtk.HBox(spacing=5)
		self.vbox.pack_start(hbox, False, True, 0)
		search_label = Gtk.Label.new_with_mnemonic(_('_Search') + ': ')
		hbox.pack_start(search_label, False, True, 0) # T: input label
		self.query_entry = InputEntry()
		self.query_entry.set_tooltip_text(HELP_TEXT)
		hbox.add(self.query_entry)
		search_label.set_mnemonic_widget(self.query_entry)
		self.search_button = Gtk.Button.new_with_mnemonic(_('_Find')) # T: Button label
		hbox.pack_start(self.search_button, False, True, 0)

		self.spinner = Gtk.Spinner()
		hbox.pack_start(self.spinner, False, True, 0)

		self.cancel_button = Gtk.Button.new_with_mnemonic(_('_Cancel')) # T: Button label
		hbox.pack_start(self.cancel_button, False, True, 0)

		self.namespacecheckbox = Gtk.CheckButton.new_with_mnemonic(_('_Limit search to the current page and sub-pages'))
			# T: checkbox option in search dialog
		if page is not None:
			self.vbox.pack_start(self.namespacecheckbox, False, True, 0)

		# TODO advanced query editor
		# TODO checkbox _('Match c_ase')
		# TODO checkbox _('Whole _word')

		self.results_treeview = SearchResultsTreeView(notebook, navigation)
		self._stack = Gtk.Stack()
		for name, widget in (
			('ready', StatusPage('edit-find-symbolic', None, HELP_TEXT)),
			('searching', StatusPage('edit-find-symbolic', _('Searching ...'))), # T: placeholder label when search has started
			('no-results', StatusPage('edit-find-symbolic', _('No results'), HELP_TEXT)), # T: placeholder label when search has no results
			('results', ScrolledWindow(self.results_treeview)),
		):
			widget.show_all()
			self._stack.add_named(widget, name)
		self.vbox.pack_start(self._stack, True, True, 0)

		self.search_button.connect_object('clicked', self.__class__._search, self)
		self.cancel_button.connect_object('clicked', self.__class__._cancel, self)
		self.query_entry.connect_object('activate', self.__class__._search, self)

		self._set_state(self.READY)

	def search(self, query):
		'''Trigger a search to be performed.
		Because search can take a long time to execute it is best to
		call this method after the dialog is shown.

		@param query: the query as string
		'''
		self.query_entry.set_text(query)
		self._search()

	def _search(self):
		string = self.query_entry.get_text()
		if self.namespacecheckbox.get_active():
			assert self.page is not None
			string = 'Section: "%s" ' % self.page.name + string
		#~ print('!! QUERY: ' + string)

		self.results_treeview.hasresults = False # XXX reset state before starting new search
		self._set_state(self.SEARCHING)
		try:
			self.results_treeview.search(string)
		except Exception as error:
			ErrorDialog(self, error).run()

		if not self.results_treeview.cancelled:
			self._set_state(self.DONE)
		else:
			self._set_state(self.CANCELLED)

	def _cancel(self):
		self.results_treeview.cancelled = True

	def _set_state(self, state):
		def hide(button):
			button.hide()
			button.set_no_show_all(True)

		def show(button):
			button.set_no_show_all(False)
			button.show_all()

		if state in (self.READY, self.DONE, self.CANCELLED):
			self.query_entry.set_sensitive(True)
			hide(self.cancel_button)
			if self.spinner:
				self.spinner.stop()
				hide(self.spinner)
			show(self.search_button)
			if state == self.READY:
				self._stack.set_visible_child_name('ready')
			elif self.results_treeview.hasresults:
				self._stack.set_visible_child_name('results')
			else:
				self._stack.set_visible_child_name('no-results')
		elif state == self.SEARCHING:
			self.query_entry.set_sensitive(False)
			hide(self.search_button)
			if self.spinner:
				show(self.spinner)
				self.spinner.start()
			show(self.cancel_button)
			if self.results_treeview.hasresults:
				self._stack.set_visible_child_name('results')
			else:
				self._stack.set_visible_child_name('searching')
		else:
			assert False, 'BUG: invalid state'



class SearchResultsTreeView(BrowserTreeView):

	NAME_COL = 0
	SCORE_COL = 1
	PATH_COL = 2

	def __init__(self, notebook, navigation):
		model = Gtk.ListStore(str, int, object)
			# NAME_COL, SCORE_COL, PATH_COL
		BrowserTreeView.__init__(self, model)
		self.navigation = navigation
		self.query = None
		self.selection = SearchSelection(notebook)
		self.cancelled = False
		self.hasresults = False

		cell_renderer = Gtk.CellRendererText()
		for name, i in (
			(_('Page'), 0), # T: Column header search dialog
			(_('Score'), 1), # T: Column header search dialog
		):
			column = Gtk.TreeViewColumn(name, cell_renderer, text=i)
			column.set_sort_column_id(i)
			if i == 0:
				column.set_expand(True)
			self.append_column(column)

		# Don't sort here because we'll do more elaborate sorting later manually#
		#model.set_sort_column_id(1, Gtk.SortType.DESCENDING)

		self.connect('row-activated', self._do_open_page)
		self.connect('destroy', self.__class__._cancel)

	def _cancel(self):
		self.cancelled = True

	def search(self, query):
		query = query.strip()
		if not query:
			return
		logger.info('Searching for: %s', query)

		self.get_model().clear()
		self.cancelled = False
		self.hasresults = False
		self.query = Query(query)
		self.selection.search(self.query, callback=self._search_callback)
		self._update_results(self.selection)

	def _search_callback(self, results, path):
		# Returning False will cancel the search
		#~ print('!! CB', path)
		if results is not None:
			self._update_results(results)

		while Gtk.events_pending():
			Gtk.main_iteration_do(False)

		return not self.cancelled

	def _update_results(self, results):
		model = self.get_model()
		if not model:
			return

		# Update score for paths that are already present
		order = []
		seen = set()
		i = -1
		for i, row in enumerate(model):
			path = row[self.PATH_COL]
			if path in results:
				score = results.scores.get(path, row[self.SCORE_COL])
			else:
				score = -1 # went missing !??? - technically a bug
			row[self.SCORE_COL] = score
			order.append((path, i, score))
			seen.add(path)

		# Add new paths
		new = results - seen
		for path in new:
			score = results.scores.get(path, 0)
			model.append((path.name, score, path))
			i += 1
			order.append((path, i, score))

		# sort by score, then by name. This doesn't seem to work by setting a sort column.
		order.sort(key=lambda i: i[0].name)
		order.sort(key=lambda i: i[2], reverse=True)
		model.reorder([x[1] for x in order])

		self.hasresults = len(model) > 0

	def _do_open_page(self, view, path, col):
		page = Path(self.get_model()[path][0])
		pageview = self.navigation.open_page(page)

		# Popup find dialog with same query
		if pageview and self.query:
			find_string, find_needs_regex = self.query.find_input
			if find_string:
				flag = FIND_REGEX if find_needs_regex else 0
				pageview.show_find(find_string, flags=flag, highlight=True)




# Copyright 2012 Jaap Karssenberg <jaap.karssenberg@gmail.com>


from gi.repository import Gtk

from zim.newfs import LocalFile
from zim.newfs.helpers import TrashHelper, TrashNotSupportedError
from zim.config import XDG_DATA_HOME, data_file
from zim.templates import list_template_categories, list_templates
from zim.gui.widgets import Dialog, BrowserTreeView, ScrolledWindow
from zim.gui.applications import open_folder_prompt_create, open_file, edit_file


class TemplateEditorDialog(Dialog):
	'''Dialog with a tree of available templates for export and new pages.
	Allows edit, delete, and create new templates. Uses external editor.
	'''

	def __init__(self, parent):
		Dialog.__init__(self, parent,
			_('Templates'), help='Help:Templates', buttons=Gtk.ButtonsType.CLOSE,
			defaultwindowsize=(400, 450))
			# T: Dialog title

		label = Gtk.Label()
		label.set_markup('<b>' + _('Templates') + '</b>')
			# T: Section in dialog
		label.set_alignment(0.0, 0.5)
		self.vbox.pack_start(label, False, True, 0)

		hbox = Gtk.HBox()
		self.vbox.pack_start(hbox, True, True, 0)

		self.view = TemplateListView()
		self.view.connect('row-activated', self.on_selection_changed)
		hbox.pack_start(ScrolledWindow(self.view), True, True, 0)

		vbbox = Gtk.VButtonBox()
		vbbox.set_layout(Gtk.ButtonBoxStyle.START)
		hbox.pack_start(vbbox, False, True, 0)

		view_button = Gtk.Button.new_with_mnemonic(_('_View')) # T: button label
		view_button.connect('clicked', self.on_view)

		copy_button = Gtk.Button.new_with_mnemonic(_('_Copy')) # T: Button label
		copy_button.connect('clicked', self.on_copy)

		edit_button = Gtk.Button.new_with_mnemonic(_('_Edit')) # T: Button label
		edit_button.connect('clicked', self.on_edit)

		delete_button = Gtk.Button.new_with_mnemonic(_('_Remove')) # T: Button label
		delete_button.connect('clicked', self.on_delete)

		for b in (view_button, copy_button, edit_button, delete_button):
			b.set_alignment(0.0, 0.5)
			vbbox.add(b)

		browse_button = Gtk.Button.new_with_mnemonic(_('Browse')) # T: button label
		browse_button.connect('clicked', self.on_browse)
		self.add_extra_button(browse_button)

		self._buttonbox = vbbox
		self._delete_button = delete_button
		self.on_selection_changed()

		## Same button appears in export dialog
		url_button = Gtk.LinkButton(
			'https://zim-wiki.org/more_templates.html',
			_('Get more templates online') # T: label for button with URL
		)
		self.vbox.pack_start(url_button, False, True, 0)

	def on_selection_changed(self, *a):
		# Set sensitivity of the buttons
		# All insensitive if category (folder) is selected
		# Delete insensitive if only a default
		custom, default = self.view.get_selected()
		for button in self._buttonbox.get_children():
			button.set_sensitive(custom is not None)

		if custom is None:
			return

		if not custom.exists():
			self._delete_button.set_sensitive(False)

	def on_view(self, *a):
		# Open the file, without waiting for editor to return
		custom, default = self.view.get_selected()
		if custom is None:
			return # Should not have been sensitive

		if custom.exists():
			open_file(self, custom)
		else:
			assert default and default.exists()
			open_file(self, default)

	def on_copy(self, *a):
		# Create a new template in this category
		custom, default = self.view.get_selected()
		if custom is None:
			return # Should not have been sensitive

		if custom.exists():
			source = custom
		else:
			assert default and default.exists()
			source = default

		name = PromptNameDialog(self).run()
		if name is None:
			return # User cancelled the operation
		_, ext = custom.basename.rsplit('.', 1)
		basename = name + '.' + ext
		newfile = custom.parent().file(basename)

		source.copyto(newfile)

		self.view.refresh()

	def on_edit(self, *a):
		custom, default = self.view.get_selected()
		if custom is None:
			return # Should not have been sensitive

		if not custom.exists():
			# Copy default
			default.copyto(custom)

		edit_file(self, custom, istextfile=True)
		self.view.refresh()

	def on_delete(self, *a):
		# Only delete custom, may result in reset to default
		custom, default = self.view.get_selected()
		if custom is None or not custom.exists():
			return # Should not have been sensitive

		try:
			TrashHelper().trash(LocalFile(custom.path))
		except TrashNotSupportedError:
			# TODO warnings
			custom.remove()

		self.view.refresh()

	def on_browse(self, *a):
		dir = XDG_DATA_HOME.folder(('zim', 'templates'))
		open_folder_prompt_create(self, dir)


class PromptNameDialog(Dialog):

	def __init__(self, parent):
		Dialog.__init__(self, parent, _('Copy Template')) # T: Dialog title
		self.add_form([
			('name', 'string', _('Name')),
				# T: Input label for the new name when copying a template
		])

	def do_response_ok(self):
		self.result = self.form['name']
		if self.result:
			return True


class TemplateListView(BrowserTreeView):

	BASENAME_COL = 0
	FILE_COL = 1
	DEFAULT_COL = 2

	def __init__(self):
		BrowserTreeView.__init__(self)
		model = Gtk.TreeStore(str, object, object)
			# BASENAME_COL, FILE_COL, DEFAULT_COL
		self.set_model(model)
		self.set_headers_visible(False)

		cell_renderer = Gtk.CellRendererText()
		column = Gtk.TreeViewColumn('_template_', cell_renderer, text=self.BASENAME_COL)
		self.append_column(column)

		self.refresh()

	def get_selected(self):
		# Returns (base, default file) or (None, None)
		model, iter = self.get_selection().get_selected()
		if model is None or iter is None:
			return None, None
		else:
			return model[iter][self.FILE_COL], model[iter][self.DEFAULT_COL]

	def select(self, path):
		self.get_selection().select_path(path)

	def refresh(self):
		model = self.get_model()
		model.clear()
		for category in list_template_categories():
			parent = model.append(None, (category, None, None))
			for name, basename in list_templates(category):
				base = XDG_DATA_HOME.file(('zim', 'templates', category, basename))
				default = data_file(('templates', category, basename)) # None if not existing
				#~ print('>>>', name, base, default)
				model.append(parent, (name, base, default))

		self.expand_all()




# Copyright 2008-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''The exporters are the main public objects for exporting, they
control the execution of the whole process. Each exporter implements a
different export style.
'''

# TODO collect errors / warnings and write a log that can be shown to the user after exporting

import logging

logger = logging.getLogger('zim.export')


class Exporter(object):
	'''Base class that defines the public API for exporter objects'''

	def export(self, pages):
		'''Export pages
		@param pages: a L{PageSelection} object
		'''
		for p in self.export_iter(pages):
			logger.info('Exporting %s', p)

	def export_iter(self, pages):
		'''Export pages while yielding page objects that are exported
		for progress monitoring

		Also note that implementations must be robust to ignore errors
		when exporting pages.

		@param pages: a L{PageSelection} object
		@implementation: must be implemented by subclasses
		'''
		raise NotImplementedError



from zim.notebook.page import Path, Page
from zim.newfs.mock import MockFile
from zim.formats import ParseTreeBuilder, \
	FORMATTEDTEXT, PARAGRAPH, HEADING, BULLETLIST, LISTITEM, LINK


def createIndexPage(notebook, path, section=None):
		# TODO make more flexible - use pages iter itself instead of section of notebook
		if section is None:
			section = Path(':')
			title = notebook.name
		else:
			title = section.name

		builder = ParseTreeBuilder()

		def add_namespace(path):
			pagelist = notebook.pages.list_pages(path)
			builder.start(BULLETLIST)
			for page in pagelist:
				builder.start(LISTITEM)
				builder.append(LINK,
					{'type': 'page', 'href': page.name}, page.basename)
				builder.text('\n')
				builder.end(LISTITEM)
				if page.haschildren:
					add_namespace(page) # recurs
			builder.end(BULLETLIST)

		builder.start(FORMATTEDTEXT)
		builder.append(HEADING, {'level': 1}, 'Index of %s\n' % title)
		builder.start(PARAGRAPH)
		add_namespace(section)
		builder.end(PARAGRAPH)
		builder.end(FORMATTEDTEXT)

		tree = builder.get_parsetree()
		#~ print("!!!", tree.tostring())

		indexpage = Page(path, False, MockFile('/index'), None, 'wiki')
		indexpage.set_parsetree(tree)
		return indexpage




# Copyright 2008-2014 Jaap Karssenberg <jaap.karssenberg@gmail.com>

'''This module defines the main parser for template text.

See also the L{expressionparser} sub-module for the parser for
expressions in the template.
'''


import re

from zim.parse.builder import BuilderTextBuffer
from zim.parse.regexparser import Rule
from zim.parse.simpletree import SimpleTreeBuilder

from zim.templates.expressionparser import ExpressionParser, ExpressionParameter


class TemplateBuilderTextBuffer(BuilderTextBuffer):
	'''Sub class of L{BuilderTextBuffer} that is used to strip text
	around template instructions.
	'''

	_rstrip_re = re.compile(r'\n?[ \t]*$')
	_lstrip_re = re.compile(r'^[ \t]*\n?')

	def __init__(self, builder):
		BuilderTextBuffer.__init__(self, builder)
		self._lstrip_pending = False

	def flush(self):
		text = ''.join(self.buffer)

		if text and self._lstrip_pending:
			text = self._lstrip(text)

		if text:
			self.builder.text(text)

		self._lstrip_pending = False
		self.buffer = []

	def rstrip(self):
		'''Do an rstrip on last piece of text up to and including
		newline
		'''
		text = ''.join(self.buffer)
		if text:
			self.buffer = [self._rstrip(text)]

	def _rstrip(self, text):
		if text.endswith('\n'):
			return text[:-1]
		else:
			return self._rstrip_re.sub('', text)

	def lstrip(self):
		'''Flag that next piece of text needs to be lstripped up to
		and including newline
		'''
		self._lstrip_pending = True

	def _lstrip(self, text):
		return self._lstrip_re.sub('', text)


class TemplateTreeBuilder(SimpleTreeBuilder):
	'''Sub-class of L{SimpleTreeBuilder}.
	This class implements a special case for the "BLOCK" element and
	always places it in the top level of the tree
	'''

	def __init__(self):
		SimpleTreeBuilder.__init__(self)

	def start(self, tag, attrib=None):
		# Special case for blocks - put them outside tree
		if tag == 'BLOCK':
			element = self.elementfactory(tag, attrib, [])
			assert self.toplevel and self.toplevel[0].tag == 'TEMPLATE'
			self.toplevel[0].append(element)
			self.stack.append(element)
		else:
			SimpleTreeBuilder.start(self, tag, attrib)


class TemplateParser(object):
	'''The parser to parse template files into a hierarchic tree of
	instructions.
	'''

	# Build parse tree like:
	#  - text
	#  - IF expr=...
	#    - text
	#    - GET expr=...
	#    - text
	#  - ELIF expr= ..
	#    - text
	#  - text
	#
	# So:
	#   GET token: append
	#   IF token: start IF
	#   ELIF token: end IF (error if unmatched), start ELIF
	#   END token: end whatever is on the stack
	#
	# Keywords are case sensitive: must be upper case

	_set_token_re = re.compile(r'^([\w.]+)\s*=\s*(.*)$') # var = expr

	_token_with_expr_re = re.compile(r'^(\w+)\s+(.*)$') # TOKEN expr

	_for_in_token_re = re.compile(r'^(\w+)\s+IN\s+(.*)$') # name IN expr

	_for_is_token_re = re.compile(r'^(\w+)\s*=\s*(.*)$') # name = expr

	_block_token_re = re.compile(r'^\w+$') # name

	_tokens_with_expr = (
		'GET', 'SET',
		'IF', 'ELIF', 'ELSIF', 'ELSE',
		'FOR', 'FOREACH',
		'BLOCK', 'INCLUDE',
		'END'
	)	# These tokens take an argument

	_tokens_without_expr = ('ELSE', 'END')
		# These tokens do not take an argument

	_tokens_with_end = (
		'IF', 'ELIF', 'ELSIF', 'ELSE',
		'FOR', 'FOREACH', 'BLOCK',
	)	# These tokens start a block that is delimited by END

	_tokens_with_line_chomp = _tokens_with_end + (
		'END', 'INCLUDE'
	)	# For these tokens strip whitespace of token is on it's own line

	_tokens_with_default_chomp = ('SET',)
		# For these tokens always strip whitespace left and right

	def __init__(self):
		self.text_parser = self.build_text_parser()
		self.expr_parser = ExpressionParser()
		self._stack = []

	def parse(self, text):
		'''Parse template text
		@param text: the template text
		@returns: a list of L{SimpleTreeElements} for the main template
		structure and any blocks defined in the template
		'''
		builder = TemplateTreeBuilder()
		self.__call__(builder, text)
		return builder.get_root()

	def __call__(self, builder, text):
		wrapper = TemplateBuilderTextBuffer(builder)
		wrapper.start('TEMPLATE')
		wrapper.start('MAIN')
		if text:
			self.text_parser(wrapper, text)
		wrapper.end('MAIN')
		wrapper.end('TEMPLATE')

	def build_text_parser(self):
		# Rules capture [% .. %] and <!--[% ... %]--> including "chomp" flags
		# First two rules block level instruction on it's own line
		# next two rules are embdedded in content
		line_tokens = '|'.join(map(re.escape, self._tokens_with_line_chomp))
		text_parser = (
			Rule('X-XML-Token', r'''
				^[^\S\n]*			# whitespace at line start
				\<\!--\[%%			# start of instruction
				(
					-?				# rchomp
					\s+
					(?:%s)			# line tokens
					(?:\s[^%%]*?)?	# optional expression -- the [^%%] os a bit of a hack here..
					\s
					-?				# lchomp
				)
				%%\]--\>			# end of instruction
				[^\S\n]*\n			# whitespace and end of line
				''' % line_tokens,
				process=self._process_token)
			| Rule('X-Text-Token', r'''
				^[^\S\n]*			# whitespace at line start
				\[%%				# start of instruction
				(
					-?				# rchomp
					\s+
					(?:%s)			# line tokens
					(?:\s[^%%]*?)?	# optional expression -- the [^%%] os a bit of a hack here..
					\s
					-?				# lchomp
				)
				%%\]				# end of instruction
				[^\S\n]*\n			# whitespace and end of line
				''' % line_tokens,
				process=self._process_token)
			| Rule('X-Inline-XML-Token',
				r'\<\!--\[%(-?\s.*?\s-?)%\]--\>',
				process=self._process_token)
			| Rule('X-Inline-Text-Token',
				r'\[%(-?\s.*?\s-?)%\]',
				process=self._process_token)
		)
		return text_parser

	def _process_token(self, builder, text):
		rchomp = text.startswith('-') # rstrip prev text
		lchomp = text.endswith('-') # lstrip next text
		text = text.strip('-').strip()

		m = self._token_with_expr_re.match(text)
		if m and m.group(1) in self._tokens_with_expr:
			token = m.group(1)
			expr = m.group(2).strip()
			assert expr, 'Missing expression' # Should never occur if regexp is strict
		elif text in self._tokens_without_expr:
			token = text
			expr = None
		else:
			m = self._set_token_re.match(text)
			if m:
				token = 'SET'
			else:
				token = 'GET'
			expr = text

		if token in self._tokens_with_default_chomp:
			rchomp = True
			lchomp = True

		if rchomp:
			builder.rstrip()

		method = getattr(self, '_process_token_' + token.lower())
		method(builder, token, expr)

		if lchomp:
			builder.lstrip()

	def _process_append_token(self, b, t, e):
		e = self.expr_parser.parse(e)
		b.append(t, {'expr': e})

	_process_token_get = _process_append_token
	_process_token_include = _process_append_token

	def _process_token_set(self, b, t, e):
		m = self._set_token_re.match(e)
		if m:
			v = ExpressionParameter(m.group(1))
			e = self.expr_parser.parse(m.group(2))
			b.append(t, {'var': v, 'expr': e})
		else:
			raise AssertionError('Invalid syntax for SET, expected assignment')

	def _process_token_if(self, b, t, e):
		e = self.expr_parser.parse(e)
		b.start('IF', {'expr': e})
		self._stack.append('IF')

	def _process_token_elif(self, b, t, e):
		if not (self._stack and self._stack[-1] in ('IF', 'ELIF')):
			raise AssertionError('Expected IF statement instead of ELIF')
		b.end(self._stack.pop()) # raises if unmatched
		e = self.expr_parser.parse(e)
		b.start('ELIF', {'expr': e})
		self._stack.append('ELIF')

	_process_token_elsif = _process_token_elif

	def _process_token_else(self, b, t, e):
		if not (self._stack and self._stack[-1] in ('IF', 'ELIF')):
			raise AssertionError('Unexpected ELSE statement')
		b.end(self._stack.pop()) # raises if unmatched
		b.start('ELSE')
		self._stack.append('ELSE')

	def _process_token_for(self, b, t, e):
		m = self._for_in_token_re.match(e)
		if t == 'FOREACH' and not m:
			m = self._for_is_token_re.match(e)

		if m:
			v = ExpressionParameter(m.group(1))
			e = self.expr_parser.parse(m.group(2))
			b.start('FOR', {'var': v, 'expr': e})
			self._stack.append('FOR')
		else:
			if t == 'FOR':
				raise AssertionError('Invalid syntax in FOR, expected "IN"')
			else:
				raise AssertionError('Invalid syntax in FOREACH, expected "IN" or "="')

	_process_token_foreach = _process_token_for

	def _process_token_block(self, b, t, e):
		if not self._block_token_re.match(e):
			raise AssertionError('Invalid syntax in BLOCK, expected name')
		b.start('BLOCK', {'name': e})
		self._stack.append('BLOCK')

	def _process_token_end(self, b, t, e):
		if not self._stack:
			raise AssertionError('Unexpected END statement')
		b.end(self._stack.pop()) # raises if unmatched



# -*- coding: UTF-8 -*-

# Copyright 2011 Jiří Janoušek <janousek.jiri@gmail.com>
# Copyright 2014-2018 Jaap Karssenberg <jaap.karssenberg@gmail.com>


from zim.signals import SignalEmitter, SIGNAL_RUN_LAST
from zim.config import String, ConfigDict


class InsertedObjectType(object):
	'''Base class for defining "objects" that can be inserted in a wiki page

	This class is called "InsertedObjectType" instead of "InsertedObject"
	because it does *not* represent a single inserted object, but defines a
	type of object of which many instances can occur. The instances themselves
	are represented by a series of tokens for the parser and a model plus a
	widget for the user interface.

	Note: if you are looking for how to define a new object type from a plugin,
	please see L{InsertedObjectTypeExtension} in L{zim.plugins}. Base classes
	for widgets can be found in L{zim.gui.insertedobjects}.
	'''

	name = None

	label = None
	verb_icon = None

	is_inline = False

	object_attr = {}

	def __init__(self):
		assert self.name is not None
		assert self.label is not None
		self.object_attr = self.object_attr.copy()
			# Prevent referencing and modifying class attribute of parent class
		self.object_attr['type'] = String(self.name)

		# each inserted object has an id attribute usable as anchor for linking
		self.object_attr['id'] = String(None)

		for name in ('model_from_data', 'data_from_model', 'format'):
			orig = getattr(self, name)
			wrapper = getattr(self, '_' + name + '_wrapper')
			setattr(self, '_inner_' + name, orig)
			setattr(self, name, wrapper)

	def parse_attrib(self, attrib):
		'''Convenience method to enforce the supported attributes and their
		types.
		@returns: a L{ConfigDict} using the C{object_attr} dict as definition
		'''
		if not isinstance(attrib, ConfigDict):
			attrib = ConfigDict(attrib)
			attrib.define(self.object_attr)
		return attrib

	def new_object(self):
		'''Create a new empty object
		@returns: a 2-tuple C{(attrib, data)}
		'''
		attrib = self.parse_attrib({})
		return attrib, ''

	def new_model_interactive(self, parent, notebook, page):
		'''Create a new object model interactively
		Interactive means that we can use e.g. a dialog to prompt for input.
		The default behavior is to use L{new_object()}.

		@param parent: Gtk widget to use as parent widget for dialogs
		@param notebook: a L{Notebook} object
		@param page: a L{Page} object for the page where this object occurs
		@returns: a model object (see L{model_from_data()})
		@raises: ValueError: if user cancelled the action
		'''
		attrib, data = self.new_object()
		return self.model_from_data(notebook, page, attrib, data)

	def _model_from_data_wrapper(self, notebook, page, attrib, data):
		attrib = self.parse_attrib(attrib)
		return self._inner_model_from_data(notebook, page, attrib, data)

	def model_from_data(self, notebook, page, attrib, data):
		'''Returns a model for the object

		The main purpose for the model is that it is shared between widgets that
		show the same object. See e.g. C{Gtk.TextBuffer} or C{Gtk.TreeModel}
		for examples.

		No API is expected of the model object other than that it can be used as
		argument for L{create_widget()} and L{data_from_model()} and a
		"changed" signal that should be emitted when the content has changed, so
		the pageview knows that the page has changed and should be saved before
		closing.

		Since the model is specific for the page where the object occurs, any
		user of the object type should serialize back to data before e.g.
		copying the object to a different page.

		This method should always be robust for missing attributes and body
		contents. The C{attrib} will automatically be checked by L{parse_attrib}
		before being given to this method.

		@param notebook: a L{Notebook} object
		@param page: a L{Page} object for the page where this object occurs
		@param attrib: dict with object attributes
		@param data: string with object content
		@returns: a model object
		'''
		raise NotImplementedError

	def _data_from_model_wrapper(self, model):
		attrib, data = self._inner_data_from_model(model)
		return attrib.copy(), data # Enforce shallow copy

	def data_from_model(self, model):
		'''Returns the object data for a model object
		This method is used to serialize the model object back into a form that
		can be handled when parsing wiki content.
		@param model: an object created with L{model_from_data()}
		@returns: a 2-tuple C{(attrib, data)}
		'''
		raise NotImplementedError

	def create_widget(self, model):
		'''Return a Gtk widget for the given model
		@param model: an object created with L{model_from_data()}
		@returns: a Gtk widget object derived from L{InsertedObjectWidget}
		'''
		raise NotImplementedError

	def _format_wrapper(self, format, dumper, attrib, data):
		attrib = self.parse_attrib(attrib)
		return self._inner_format(format, dumper, attrib, data)

	def format(self, format, dumper, attrib, data):
		'''Format the object using a specific output format
		Intended to improve rendering of the object on exporting.

		This method should always be robust for missing attributes and body
		contents. The C{attrib} will automatically be checked by L{parse_attrib}
		before being given to this method.

		Implementing this method is optional, default checks for a specific
		method per format (e.g. C{format_html()} for the "html" format) and
		raises C{ValueError} if no such method is defined.

		@param format: name of the output format
		@param dumper: L{Dumper} object
		@param attrib: dict with object attributes
		@param data: string with object content
		@returns: a list of strings
		@raises ValueError: if no specific formatting for "format" is available
		'''
		try:
			method = getattr(self, 'format_' + format)
		except AttributeError:
			raise ValueError('No "%s" formatting defined for objecttype "%s"' % (format, self.name))
		else:
			return method(dumper, attrib, data)
